"""
Implementation of a class for the analysis of hyperfine structure spectra.

.. moduleauthor:: Wouter Gins <wouter.gins@kuleuven.be>
.. moduleauthor:: Ruben de Groote <ruben.degroote@kuleuven.be>
"""
import copy
from fractions import Fraction

import lmfit as lm
from satlas.models.basemodel import BaseModel, SATLASParameters
from satlas.utilities import poisson_interval
import satlas.profiles as p
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy.optimize as optimize
from sympy.physics.wigner import wigner_6j, wigner_3j

W6J = wigner_6j
W3J = wigner_3j

__all__ = ['HFSModel']


class HFSModel(BaseModel):

    r"""Constructs a HFS spectrum, consisting of different
    peaks described by a certain profile. The number of peaks is governed by
    the nuclear spin and the atomic spins of the levels."""

    __shapes__ = {'gaussian': p.Gaussian,
                  'lorentzian': p.Lorentzian,
                  'crystalball': p.Crystalball,
                  'voigt': p.Voigt,
                  'pseudovoigt': p.PseudoVoigt,
                  'asymmlorentzian': p.AsymmLorentzian}

    def __init__(self, I, J, ABC, centroid, fwhm=[50.0, 50.0], scale=1.0, background_params=[0.001], shape='voigt', use_racah=False, use_saturation=False, saturation=0.001, shared_fwhm=True, sidepeak_params={'N': 0, 'Poisson': 0.68, 'Offset': 0}, crystalballparams={'Taillocation': 1, 'Tailamplitude': 1}, pseudovoigtparams={'Eta': 0.5, 'A': 0}, asymmetryparams={'a': 0}):
        """Builds the HFS with the given atomic and nuclear information.

        Parameters
        ----------
        I: float
            The nuclear spin.
        J: list of 2 floats
            The spins of the fine structure levels.
        ABC: list of 6 floats
            The hyperfine structure constants A, B and C for ground- and excited
            fine level. The list should be given as [A :sub:`lower`,
            A :sub:`upper`, B :sub:`lower`, B :sub:`upper`, C :sub:`upper`,
            C :sub:`lower`].
        centroid: float
            Centroid of the spectrum.

        Other parameters
        ----------------
        fwhm: float or list of 2 floats, optional
            Depending on the used shape, the FWHM is defined by one or two floats.
            Defaults to [50.0, 50.0]
        scale: float, optional
            Sets the strength of the spectrum, defaults to 1.0. Comparable to the
            amplitude of the spectrum.
        background_params: list of float, optional
            Sets the coefficients of the polynomial background to the given values.
            Order of polynomial is equal to the number of parameters given minus one.
            Highest order coefficient is the first element, etc.
        shape : string, optional
            Sets the transition shape. String is converted to lowercase. For
            possible values, see *HFSModel__shapes__*.keys()`.
            Defaults to Voigt if an incorrect value is supplied.
        use_racah: boolean, optional
            If True, fixes the relative peak intensities to the Racah intensities.
            Otherwise, gives them equal intensities and allows them to vary during
            fitting.
        use_saturation: boolean, optional
            If True, uses the saturation parameter to calculate relative intensities.
        saturation: float, optional
            If different than 0, calculate the saturation effect on the intensity of
            transition intensity. This is done by an exponential transition between
            Racah intensities and the saturated intensities.
        shared_fwhm: boolean, optional
            If True, the same FWHM is used for all peaks. Otherwise, give them all
            the same initial FWHM and let them vary during the fitting.
        sidepeak_params: dict
            A dictionary with the following keys and values:

            n: int
                Sets the number of sidepeaks that are present in the spectrum.
                Defaults to 0.
            poisson: float
                Sets the relative intensity of the first side peak. The intensity of the
                other sidepeaks is calculated from the Poisson-factor.
            offset: float
                Sets the distance (in MHz) of each sidepeak in the spectrum.
        crystalballparams: dict
            A dictionary with the following keys and values:

            tailamp: float
                Sets the relative amplitude of the tail for the Crystalball shape function.
            tailloc: float
                Sets the location of the tail for the Crystalball shape function.
        pseudovoigtparams: dict
            A dictionary with the following keys and values:

            Eta: float between 0 and 1
                Describes the mixing percentage of the Gaussian and Lorentzian shapes
            A: float
                Describes the asymmetry of the peak.

        Note
        ----
        The list of parameter keys is:
            * *FWHM* (only for profiles with one float for the FWHM)
            * *FWHMG* (only for profiles with two floats for the FWHM)
            * *FWHML* (only for profiles with two floats for the FWHM)
            * *Al*
            * *Au*
            * *Bl*
            * *Bu*
            * *Cl*
            * *Cu*
            * *Centroid*
            * *Background*
            * *Poisson* (only if the attribute *n* is greater than 0)
            * *Offset* (only if the attribute *n* is greater than 0)
            * *Amp* (with the correct labeling of the transition)
            * *scale*"""
        super(HFSModel, self).__init__()
        shape = shape.lower()
        if shape not in self.__shapes__:
            print("""Given profile shape not yet supported.
            Defaulting to Voigt lineshape.""")
            shape = 'voigt'
            fwhm = [50.0, 50.0]

        self.I_value = {0.0: ((False, 0), (False, 0), (False, 0),
                              (False, 0), (False, 0), (False, 0)),
                        0.5: ((True, 1), (True, 1),
                              (False, 0), (False, 0), (False, 0), (False, 0)),
                        1.0: ((True, 1), (True, 1),
                              (True, 1), (True, 1),
                              (False, 0), (False, 0))
                        }
        self.J_lower_value = {0.0: ((False, 0), (False, 0), (False, 0)),
                              0.5: ((True, 1),
                                    (False, 0), (False, 0)),
                              1.0: ((True, 1),
                                    (True, 1), (False, 0))
                              }
        self.J_upper_value = {0.0: ((False, 0), (False, 0), (False, 0)),
                              0.5: ((True, 1),
                                    (False, 0), (False, 0)),
                              1.0: ((True, 1),
                                    (True, 1), (False, 0))
                              }
        self.shape = shape
        self._use_racah = use_racah
        self._use_saturation = use_saturation
        self.shared_fwhm = shared_fwhm
        self.I = I
        self.J = J
        self._calculate_F_levels()
        self._calculate_energy_coefficients()
        self._calculate_transitions()

        self._vary = {}
        self._constraints = {}

        self.ratioA = (None, 'lower')
        self.ratioB = (None, 'lower')
        self.ratioC = (None, 'lower')

        self._roi = (-np.inf, np.inf)

        self._populateparams(ABC, centroid, fwhm, scale, saturation,
                              background_params,
                              sidepeak_params,
                              crystalballparams,
                              pseudovoigtparams,
                              asymmetryparams)

    @property
    def locations(self):
        """Contains the locations of the peaks."""
        return self._locations

    @locations.setter
    def locations(self, locations):
        self._locations = np.array(locations)
        for p, l in zip(self.parts, locations):
            p.mu = l

    @property
    def use_racah(self):
        """Boolean to set the behaviour to Racah intensities (True)
        or to individual amplitudes (False)."""
        return self._use_racah

    @use_racah.setter
    def use_racah(self, value):
        self._use_racah = value
        self.params['Scale'].vary = self._use_racah or self._use_saturation
        for label in self.ftof:
            self.params['Amp' + label].vary = not (self._use_racah or self._use_saturation)

    @property
    def use_saturation(self):
        """Boolean to set the behaviour to the saturation model (True)
        or not (False)."""
        return self._use_saturation

    @use_saturation.setter
    def use_saturation(self, value):
        self._use_saturation = value
        self.params['Saturation'].vary = value
        self.params['Scale'].vary = self._use_racah or self._use_saturation
        for label in self.ftof:
            self.params['Amp' + label].vary = not (self._use_racah or self._use_saturation)

    @property
    def params(self):
        """Instance of lmfit.Parameters object characterizing the
        shape of the HFS."""
        return self._parameters

    @params.setter
    def params(self, params):
        p = params.copy()
        p._prefix = self._prefix
        self._parameters = self._check_variation(p)
        # self._parameters.pretty_print()
        # print(self._prefix)
        # When changing the parameters, the energies and
        # the locations have to be recalculated
        self._calculate_energies()
        self._calculate_transition_locations()
        if not self.use_racah and not self.use_saturation:
            # When not using set amplitudes, they need
            # to be changed after every iteration
            self._set_amplitudes()
        elif self.use_saturation:
            self._set_transitional_amplitudes()
        else:
            pass
        # Finally, the fwhm of each peak needs to be set
        self._set_fwhm()
        if self.shape.lower() == 'crystalball':
            for part in self.parts:
                part.alpha = self.params['Taillocation'].value
                part.n = self.params['Tailamplitude'].value
        if self.shape.lower() == 'pseudovoigt':
            for label, part in zip(self.ftof, self.parts):
                if self.shared_fwhm:
                    part.n = self.params['Eta'].value
                    part.a = self.params['Asym'].value
                else:
                    part.n = self.params['Eta'+label].value
                    part.a = self.params['Asym'+label].value
        elif self.shape.lower() == 'asymmlorentzian':
            for label, part in zip(self.ftof, self.parts):
                if self.shared_fwhm:
                    part.asymm = self.params['Asym'].value
                else:
                    part.asymm = self.params['Asym'+label].value

    def _set_transitional_amplitudes(self):
        values = self._calculate_transitional_intensities(self.params['Saturation'].value)
        for p, l, v in zip(self.parts, self.ftof, values):
            self.params['Amp' + l].value = v
            p.amp = v

    @property
    def ftof(self):
        """List of transition labels, of the form *Flow__Fhigh* (half-integers
        have an underscore instead of a division sign), same ordering
        as given by the attribute :attr:`.locations`."""
        return self._ftof

    @ftof.setter
    def ftof(self, value):
        self._ftof = value

    def _calculate_energies(self):
        r"""The hyperfine addition to a central frequency (attribute *centroid*)
        for a specific level is calculated. The formula comes from
        :cite:`Schwartz1955` and in a simplified form, reads

        .. math::
            C_F &= F(F+1) - I(I+1) - J(J+1)

            D_F &= \frac{3 C_F (C_F + 1) - 4 I (I + 1) J (J + 1)}{2 I (2 I - 1)
            J (2 J - 1)}

            E_F &= \frac{10 (\frac{C_F}{2})^3 + 20(\frac{C_F}{2})^2 + C_F(-3I(I
            + 1)J(J + 1) + I(I + 1) + J(J + 1) + 3) - 5I(I + 1)J(J + 1)}{I(I -
            1)(2I - 1)J(J - 1)(2J - 1)}

            E &= centroid + \frac{A C_F}{2} + \frac{B D_F}{4} + C E_F

        A, B and C are the dipole, quadrupole and octupole hyperfine
        parameters. Octupole contributions are calculated when both the
        nuclear and electronic spin is greater than 1, quadrupole contributions
        when they are greater than 1/2, and dipole contributions when they are
        greater than 0.

        Parameters
        ----------
        level: int, 0 or 1
            Integer referring to the lower (0) level, or the upper (1) level.
        F: integer or half-integer
            F-quantum number for which the hyperfine-corrected energy has to be
            calculated.

        Returns
        -------
        energy: float
            Energy in MHz."""
        A = np.append(np.ones(self.num_lower) * self.params['Al'].value,
                      np.ones(self.num_upper) * self.params['Au'].value)
        B = np.append(np.ones(self.num_lower) * self.params['Bl'].value,
                      np.ones(self.num_upper) * self.params['Bu'].value)
        C = np.append(np.ones(self.num_lower) * self.params['Cl'].value,
                      np.ones(self.num_upper) * self.params['Cu'].value)
        centr = np.append(np.zeros(self.num_lower),
                          np.ones(self.num_upper) * self.params['Centroid'].value)
        self.energies = centr + self.C * A + self.D * B + self.E * C

    def _calculate_transition_locations(self):
        self.locations = [self.energies[ind_high] - self.energies[ind_low] for (ind_low, ind_high) in self.transition_indices]

    def _set_amplitudes(self):
        for p, label in zip(self.parts, self.ftof):
            p.amp = self.params['Amp' + label].value

    def _set_fwhm(self):
        if self.shape.lower() == 'voigt':
            fwhm = [[self.params['FWHMG'].value, self.params['FWHML'].value] for _ in self.ftof] if self.shared_fwhm else [[self.params['FWHMG' + label].value, self.params['FWHML' + label].value] for label in self.ftof]
        else:
            fwhm = [self.params['FWHM'].value for _ in self.ftof] if self.shared_fwhm else [self.params['FWHM' + label].value for label in self.ftof]
        for p, f in zip(self.parts, fwhm):
            p.fwhm = f

    ####################################
    #      INITIALIZATION METHODS      #
    ####################################

    def _populateparams(self, ABC, centroid, fwhm, scale, saturation, background_params, sidepeak_params, crystalballparams, pseudovoigtparams, asymmetryparams):
        # Prepares the params attribute with the initial values
        par = SATLASParameters()
        if not self.shape.lower() == 'voigt':
            if self.shared_fwhm:
                par.add('FWHM', value=fwhm, vary=True, min=0)
                if self.shape.lower() == 'pseudovoigt':
                    Eta = pseudovoigtparams['Eta']
                    A = pseudovoigtparams['A']
                    par.add('Eta', value=Eta, vary=True, min=0, max=1)
                    par.add('Asym', value=A, vary=True)
                if self.shape.lower() == 'asymmlorentzian':
                    par.add('Asym', value=asymmetryparams['a'], vary=True)
            else:
                fwhm = [fwhm for _ in range(len(self.ftof))]
                for label, val in zip(self.ftof, fwhm):
                    par.add('FWHM' + label, value=val, vary=True, min=0)
                    if self.shape.lower() == 'pseudovoigt':
                        Eta = pseudovoigtparams['Eta']
                        A = pseudovoigtparams['A']
                        par.add('Eta' + label, value=Eta, vary=True, min=0, max=1)
                        par.add('Asym' + label, value=A, vary=True)
                    if self.shape.lower() == 'asymmlorentzian':
                        par.add('Asym' + label, value=asymmetryparams['a'], vary=True)
        else:
            if self.shared_fwhm:
                par.add('FWHMG', value=fwhm[0], vary=True, min=1)
                par.add('FWHML', value=fwhm[1], vary=True, min=1)
                val = 0.5346 * fwhm[1] + np.sqrt(0.2166 * fwhm[1] ** 2 + fwhm[0] ** 2)
                par.add('TotalFWHM', value=val, vary=False,
                        expr='0.5346*FWHML+(0.2166*FWHML**2+FWHMG**2)**0.5')
            else:
                fwhm = np.array(fwhm)
                fwhm = np.array([[fwhm[0], fwhm[1]] for _ in range(len(self.ftof))])
                for label, val in zip(self.ftof, fwhm):
                    par.add('FWHMG' + label, value=val[0], vary=True, min=0)
                    par.add('FWHML' + label, value=val[1], vary=True, min=0)
                    val = 0.5346 * val[1] + np.sqrt(0.2166 * val[1] ** 2
                                                    + val[0] ** 2)
                    par.add('TotalFWHM' + label, value=val, vary=False,
                            expr='0.5346*FWHML' + label +
                                 '+(0.2166*FWHML' + label +
                                 '**2+FWHMG' + label + '**2)**0.5')
        if self.shape.lower() == 'crystalball':
            taillocation = crystalballparams['Taillocation']
            tailamplitude = crystalballparams['Tailamplitude']
            par.add('Taillocation', value=taillocation, vary=True)
            par.add('Tailamplitude', value=tailamplitude, vary=True)
            for part in self.parts:
                part.alpha = taillocation
                part.n = tailamplitude

        par.add('Scale', value=scale, vary=self.use_racah or self.use_saturation, min=0)
        par.add('Saturation', value=saturation * self.use_saturation, vary=self.use_saturation, min=0)
        amps = self._calculate_transitional_intensities(saturation)
        for label, amp in zip(self.ftof, amps):
            label = 'Amp' + label
            par.add(label, value=amp, vary=not (self.use_racah or self.use_saturation), min=0)

        par.add('Al', value=ABC[0], vary=True)
        par.add('Au', value=ABC[1], vary=True)
        par.add('Bl', value=ABC[2], vary=True)
        par.add('Bu', value=ABC[3], vary=True)
        par.add('Cl', value=ABC[4], vary=True)
        par.add('Cu', value=ABC[5], vary=True)

        ratios = (self.ratioA, self.ratioB, self.ratioC)
        labels = (('Al', 'Au'), ('Bl', 'Bu'), ('Cl', 'Cu'))
        for r, (l, u) in zip(ratios, labels):
            if r[0] is not None:
                if r[1].lower() == 'lower':
                    fixed, free = l, u
                else:
                    fixed, free = u, l
                par[fixed].expr = str(r[0]) + '*' + free
                par[fixed].vary = False

        par.add('Centroid', value=centroid, vary=True)

        for i, val in enumerate(reversed(background_params)):
            par.add('Background' + str(i), value=background_params[i], vary=True)
        self.background_degree = i
        n, poisson, offset = sidepeak_params['N'], sidepeak_params['Poisson'], sidepeak_params['Offset']
        par.add('N', value=n, vary=False)
        if n > 0:
            par.add('Poisson', value=poisson, vary=True, min=0, max=1)
            par.add('Offset', value=offset, vary=False, min=None, max=None)

        self.params = self._check_variation(par)

    def _set_ratios(self, par):
        # Process the set ratio's for the hyperfine parameters.
        ratios = (self.ratioA, self.ratioB, self.ratioC)
        labels = (('Al', 'Au'), ('Bl', 'Bu'), ('Cl', 'Cu'))
        for r, (l, u) in zip(ratios, labels):
            if r[0] is not None:
                if r[1].lower() == 'lower':
                    fixed, free = l, u
                else:
                    fixed, free = u, l
                par[fixed].expr = str(r[0]) + '*' + free
                par[fixed].vary = False
                par[free].vary = True
        return par

    def _check_variation(self, par):
        par = super(HFSModel, self)._check_variation(par)
        # Make sure the variations in the params are set correctly.
        for key in self._vary.keys():
            if key in par.keys():
                par[key].vary = self._vary[key]
        par['N'].vary = False

        if self.I in self.I_value:
            Al, Au, Bl, Bu, Cl, Cu = self.I_value[self.I]
            if not Al[0]:
                par['Al'].vary, par['Al'].value = Al
            if not Au[0]:
                par['Au'].vary, par['Au'].value = Au
            if not Bl[0]:
                par['Bl'].vary, par['Bl'].value = Bl
            if not Bu[0]:
                par['Bu'].vary, par['Bu'].value = Bu
            if not Cl[0]:
                par['Cl'].vary, par['Cl'].value = Cl
            if not Cu[0]:
                par['Cu'].vary, par['Cu'].value = Cu
        if self.J[0] in self.J_lower_value:
            Al, Bl, Cl = self.J_lower_value[self.J[0]]
            if not Al[0]:
                par['Al'].vary, par['Al'].value = Al
            if not Bl[0]:
                par['Bl'].vary, par['Bl'].value = Bl
            if not Cl[0]:
                par['Cl'].vary, par['Cl'].value = Cl
        if self.J[self.num_lower] in self.J_upper_value:
            Au, Bu, Cu = self.J_upper_value[self.J[self.num_lower]]
            if not Au[0]:
                par['Au'].vary, par['Au'].value = Au
            if not Bu[0]:
                par['Bu'].vary, par['Bu'].value = Bu
            if not Cu[0]:
                par['Cu'].vary, par['Cu'].value = Cu

        for key in self._constraints.keys():
            for bound in self._constraints[key]:
                if bound.lower() == 'min':
                    par[key].min = self._constraints[key][bound]
                elif bound.lower() == 'max':
                    par[key].max = self._constraints[key][bound]
                else:
                    pass
        return par

    def _calculate_F_levels(self):
        F1 = np.arange(abs(self.I - self.J[0]), self.I+self.J[0]+1, 1)
        self.num_lower = len(F1)
        F2 = np.arange(abs(self.I - self.J[1]), self.I+self.J[1]+1, 1)
        self.num_upper = len(F2)
        F = np.append(F1, F2)
        self.J = np.append(np.ones(len(F1)) * self.J[0],
                           np.ones(len(F2)) * self.J[1])
        self.F = F

    def _calculate_transitions(self):
        f_f = []
        indices = []
        amps = []
        sat_amp = []
        for i, F1 in enumerate(self.F[:self.num_lower]):
            for j, F2 in enumerate(self.F[self.num_lower:]):
                if abs(F2 - F1) <= 1 and not F2 == F1 == 0.0:
                    sat_amp.append(2*F1+1)
                    j += self.num_lower
                    intensity = self._calculate_racah_intensity(float(self.J[i]),
                                                               float(self.J[j]),
                                                               float(self.F[i]),
                                                               float(self.F[j]))
                    if intensity > 0:
                        amps.append(intensity)
                        indices.append([i, j])
                        s = ''
                        temp = Fraction(F1).limit_denominator()
                        if temp.denominator == 1:
                            s += str(temp.numerator)
                        else:
                            s += str(temp.numerator) + '_' + str(temp.denominator)
                        s += '__'
                        temp = Fraction(F2).limit_denominator()
                        if temp.denominator == 1:
                            s += str(temp.numerator)
                        else:
                            s += str(temp.numerator) + '_' + str(temp.denominator)
                        f_f.append(s)
        self.ftof = f_f  # Stores the labels of all transitions, in order
        self.transition_indices = indices  # Stores the indices in the F and energy arrays for the transition

        self.racah_amplitudes = np.array(amps)  # Sets the initial amplitudes to the Racah intensities
        self.racah_amplitudes = self.racah_amplitudes / self.racah_amplitudes.max()

        self.saturated_amplitudes = np.array(sat_amp)
        self.saturated_amplitudes = self.saturated_amplitudes / self.saturated_amplitudes.max()

        self.parts = tuple(self.__shapes__[self.shape](amp=a) for a in self.racah_amplitudes)

    def _calculate_transitional_intensities(self, s):
        if s <= 0:
            return self.racah_amplitudes
        else:
            sat = self.saturated_amplitudes
            rac = self.racah_amplitudes
            transitional = -sat*np.expm1(-rac * s / sat)
            return transitional / transitional.max()

    def _calculate_racah_intensity(self, J1, J2, F1, F2, order=1.0):
        return float((2 * F1 + 1) * (2 * F2 + 1) * \
                     W6J(J2, F2, self.I, F1, J1, order) ** 2)  # DO NOT REMOVE CAST TO FLOAT!!!

    def _calculate_energy_coefficients(self):
        # Since I, J and F do not change, these factors can be calculated once
        # and then stored.
        I, J, F = self.I, self.J, self.F
        C = (F*(F+1) - I*(I+1) - J*(J + 1)) * (J/J) if I > 0 else 0 * J  #*(J/J) is a dirty trick to avoid checking for J=0
        D = (3*C*(C+1) - 4*I*(I+1)*J*(J+1)) / (2*I*(2*I-1)*J*(2*J-1))
        E = (10*(0.5*C)**3 + 20*(0.5*C)**2 + C*(-3*I*(I+1)*J*(J+1) + I*(I+1) + J*(J+1) + 3) - 5*I*(I+1)*J*(J+1)) / (I*(I-1)*(2*I-1)*J*(J-1)*(2*J-1))
        C = np.where(np.isfinite(C), 0.5 * C, 0)
        D = np.where(np.isfinite(D), 0.25 * D, 0)
        E = np.where(np.isfinite(E), E, 0)
        self.C, self.D, self.E = C, D, E

    ##########################
    #      USER METHODS      #
    ##########################
    def fix_ratio(self, value, target='upper', parameter='A'):
        """Fixes the ratio for a given hyperfine parameter to the given value.

        Parameters
        ----------
        value: float
            Value to which the ratio is set
        target: {'upper', 'lower'}
            Sets the target level. If 'upper', the upper parameter is
            calculated as lower * ratio, 'lower' calculates the lower
            parameter as upper * ratio.
        parameter: {'A', 'B', 'C'}
            Selects which hyperfine parameter to set the ratio for."""
        if target.lower() not in ['lower', 'upper']:
            raise KeyError("Target must be 'lower' or 'upper'.")
        if parameter.lower() not in ['a', 'b', 'c']:
            raise KeyError("Parameter must be 'A', 'B' or 'C'.")
        if parameter.lower() == 'a':
            self.ratioA = (value, target)
        if parameter.lower() == 'b':
            self.ratioB = (value, target)
        if parameter.lower() == 'c':
            self.ratioC = (value, target)
        self.params = self._set_ratios(self.params)

    ###########################
    #      MAGIC METHODS      #
    ###########################

    def __call__(self, x):
        """Get the response for frequency *x* (in MHz) of the spectrum.

        Parameters
        ----------
        x : float or array_like
            Frequency in MHz

        Returns
        -------
        float or NumPy array
            Response of the spectrum for each value of *x*."""
        if self.params['N'].value > 0:
            s = np.zeros(x.shape)
            for i in range(self.params['N'].value + 1):
                s += (self.params['Poisson'].value ** i) * (sum([prof(x - i * self.params['Offset'].value)
                                                                for prof in self.parts])) / np.math.factorial(i)
            s *= self.params['Scale'].value
        else:
            s = self.params['Scale'].value * sum([prof(x) for prof in self.parts])
        # background_params = [self.params[par_name].value for par_name in self.params if par_name.startswith('Background')]
        background_params = [self.params['Background' + str(int(deg))].value for deg in reversed(list(range(self.background_degree + 1)))]
        return s + np.polyval(background_params, x)

    ###############################
    #      PLOTTING ROUTINES      #
    ###############################

    def plot(self, x=None, y=None, yerr=None,
             no_of_points=10**3, ax=None, show=True, plot_kws={}):
        """Plot the hfs, possibly on top of experimental data.

        Parameters
        ----------
        x: array
            Experimental x-data. If None, a suitable region around
            the peaks is chosen to plot the hfs.
        y: array
            Experimental y-data.
        yerr: array or dict('high': array, 'low': array)
            Experimental errors on y.
        no_of_points: int
            Number of points to use for the plot of the hfs if
            experimental data is given.
        ax: matplotlib axes object
            If provided, plots on this axis.
        show: boolean
            If True, the plot will be shown at the end.
        plot_kws: dictionary
            A dictionary possibly containing the following entries:

            legend: string, optional
                If given, an entry in the legend will be made for the spectrum.
            data_legend: string, optional
                If given, an entry in the legend will be made for the experimental
                data.
            xlabel: string, optional
                If given, sets the xlabel to this string. Defaults to 'Frequency (MHz)'.
            ylabel: string, optional
                If given, sets the ylabel to this string. Defaults to 'Counts'.
            model: boolean, optional
                If given, the region around the fitted line will be shaded, with
                the luminosity indicating the pmf of the Poisson
                distribution characterized by the value of the fit. Note that
                the argument *yerr* is ignored if *model* is True.
            normalized: boolean, optional
                If True, the data and fit are plotted normalized such that the highest
                data point is one.
            background: boolean, optional
                If True, the background is used, otherwise the pure spectrum is plotted.

        Returns
        -------
        fig, ax: matplotlib figure and axis
            Figure and axis used for the plotting."""

        kws = copy.deepcopy(plot_kws)
        legend = kws.pop('legend', None,)
        data_legend = kws.pop('data_legend', None)
        xlabel = kws.pop('xlabel', 'Frequency (MHz)')
        ylabel = kws.pop('ylabel', 'Counts',)
        indicate = kws.pop('indicate', False)
        model = kws.pop('model', False)
        colormap = kws.pop('colormap', 'bone_r',)
        normalized = kws.pop('normalized', False)
        distance = kws.pop('distance', 4)
        background = kws.pop('background', True)

        if ax is None:
            fig, ax = plt.subplots(1, 1)
        else:
            fig = ax.get_figure()
        toReturn = fig, ax
        color_points = next(ax._get_lines.prop_cycler)['color']
        color_lines = next(ax._get_lines.prop_cycler)['color']

        if x is None:
            ranges = []
            fwhm = self.parts[0].fwhm

            for pos in self.locations:
                r = np.linspace(pos - distance * fwhm,
                                pos + distance * fwhm,
                                2 * 10**2*0+50)
                ranges.append(r)
            superx = np.sort(np.concatenate(ranges))
        else:
            superx = np.linspace(x.min(), x.max(), int(no_of_points))

        if 'sigma_x' in self.params:
            xerr = self.params['sigma_x'].value
        else:
            xerr = 0

        if normalized:
            norm = np.max(y)
            y,yerr = y/norm,yerr/norm
        else:
            norm = 1

        if x is not None and y is not None:
            if not model:
                try:
                    ax.errorbar(x, y, yerr=[yerr['low'], yerr['high']],
                                xerr=xerr, fmt='o', label=data_legend, color=color_points)
                except:
                    ax.errorbar(x, y, yerr=yerr, fmt='o', label=data_legend, color=color_points)
            else:
                ax.plot(x, y, 'o', color=color_points)
        if model:
            superx = np.linspace(superx.min(), superx.max(), len(superx))
            range = (self.locations.min(), self.locations.max())
            if range[0] == range[1]:
                max_counts = self(range[0])
            else:
                max_counts = np.ceil(-optimize.brute(lambda x: -self(x), (range,), full_output=True, Ns=1000, finish=optimize.fmin)[1])
            min_counts = [self.params[par_name].value for par_name in self.params if par_name.startswith('Background')][-1]
            min_counts = np.floor(max(0, min_counts - 3 * min_counts ** 0.5))
            y = np.arange(min_counts, max_counts + 3 * max_counts ** 0.5 + 1)
            x, y = np.meshgrid(superx, y)
            from scipy import stats
            z = stats.poisson(self(x)).pmf(y)

            z = z / z.sum(axis=0)
            ax.imshow(z, extent=(x.min(), x.max(), y.min(), y.max()), cmap=plt.get_cmap(colormap))
            line, = ax.plot(superx, self(superx) / norm, label=legend, lw=0.5, color=color_lines)
        else:
            if background:
                y = self(superx)
            else:
                background_params = [self.params[par_name].value for par_name in self.params if par_name.startswith('Background')]
                y = self(superx) - np.polyval(background_params, superx)
            line, = ax.plot(superx, y, label=legend, color=color_lines)
        ax.set_xlim(superx.min(), superx.max())
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        if show:
            plt.show()
        return toReturn

    def plot_spectroscopic(self, **kwargs):
        """Plots the hfs on top of experimental data
        with errorbar given by the square root of the data.

        Parameters
        ----------
        x: array
            Experimental x-data. If None, a suitable region around
            the peaks is chosen to plot the hfs.
        y: array
            Experimental y-data.
        yerr: array or dict('high': array, 'low': array)
            Experimental errors on y.
        no_of_points: int
            Number of points to use for the plot of the hfs if
            experimental data is given.
        ax: matplotlib axes object
            If provided, plots on this axis.
        show: boolean
            If True, the plot will be shown at the end.
        legend: string, optional
            If given, an entry in the legend will be made for the spectrum.
        data_legend: string, optional
            If given, an entry in the legend will be made for the experimental
            data.

        Returns
        -------
        fig, ax: matplotlib figure and axis
            Figure and axis used for the plotting."""
        y = kwargs.get('y', None)
        if y is not None:
            ylow, yhigh = poisson_interval(y)
            yerr = {'low': y - ylow, 'high': yhigh - y}
        else:
            yerr = None
        kwargs['yerr'] = yerr
        return self.plot(**kwargs)


    def plot_scheme(self, show=True, upper_color='#D55E00', lower_color='#009E73', arrow_color='#0072B2', distance=5, spectrum=False):
        """Create a figure where both the splitting of the upper and lower state is drawn,
        and the hfs associated with this.

        Parameters
        ----------
        show: boolean, optional
            If True, immediately shows the figure. Defaults to True.
        upper_color: matplotlib color definition
            Sets the color of the upper state. Defaults to red.
        lower_color: matplotlib color definition
            Sets the color of the lower state. Defaults to black.
        arrow_color: matplotlib color definition
            Sets the color of the arrows indicating the transitions.
            Defaults to blue.

        Returns
        -------
        tuple
            Tuple containing the figure and both axes, also in a tuple."""
        from fractions import Fraction
        from matplotlib import lines
        length_plot = 0.4
        fig = plt.figure(frameon=False)
        ax = fig.add_axes([0.5, 0, length_plot, 0.5], axisbg=[1, 1, 1, 0])
        self.plot(ax=ax, show=False, plot_kws={'distance': distance})
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)

        locations = self.locations
        plotrange = ax.get_xlim()
        distances = (locations - plotrange[0]) / (plotrange[1] - plotrange[0]) * length_plot
        height = self(locations)
        plotrange = ax.get_ylim()
        if not spectrum:
            plt.close(fig)
            fig = plt.figure(frameon=False)
            lower_state_height = 0.1
        else:
            lower_state_height = 0.525
        upper_state_height = 0.95
        height = (height - plotrange[0]) / (plotrange[1] - plotrange[0]) / 2
        A = np.append(np.ones(self.num_lower) * self.params['Al'].value,
                              np.ones(self.num_upper) * self.params['Au'].value)
        B = np.append(np.ones(self.num_lower) * self.params['Bl'].value,
                              np.ones(self.num_upper) * self.params['Bu'].value)
        C = np.append(np.ones(self.num_lower) * self.params['Cl'].value,
                              np.ones(self.num_upper) * self.params['Cu'].value)
        energies = self.C * A + self.D * B + self.E * C

        energy_range = (upper_state_height - lower_state_height) / 2 - 0.025

        energies_upper = energies[self.num_lower:]
        energies_upper_norm = np.abs(energies_upper.max()) if not energies_upper.max()==0 else 1
        energies_upper_norm = np.ptp(energies_upper)
        energies_upper = energies_upper / energies_upper_norm * energy_range
        energies_lower = energies[:self.num_lower]
        energies_lower_norm = np.abs(energies_lower.max()) if not energies_lower.max()==0 else 1
        energies_lower_norm = np.ptp(energies_lower)
        energies_lower = energies_lower / energies_lower_norm * energy_range
        lower_state_height -= energies_lower.min()
        upper_state_height -= energies_upper.max()
        energies = np.append(energies_lower, energies_upper)

        ax2 = fig.add_axes([0, 0, 1, 1], axisbg=[1, 1, 1, 0])
        ax2.get_xaxis().set_visible(False)
        ax2.get_yaxis().set_visible(False)
        ax2.axis('off')

        # Lower state
        x = np.array([0, 0.3])
        y = np.array([lower_state_height, lower_state_height])
        line = lines.Line2D(x, y, lw=2., color=lower_color)
        ax2.add_line(line)

        # Upper state
        x = np.array([0, 0.3])
        y = np.array([upper_state_height, upper_state_height])
        line = lines.Line2D(x, y, lw=2., color=upper_color)
        ax2.add_line(line)

        for i, F in enumerate(self.F):
            # Level
            F = Fraction.from_float(F)
            x = np.array([0.5, 0.5 + length_plot])
            if i < self.num_lower:
                y = np.zeros(len(x)) + lower_state_height + energies[i]
                color = lower_color
                starting = lower_state_height
            else:
                y = np.zeros(len(x)) + upper_state_height + energies[i]
                color = upper_color
                starting = upper_state_height
            line = lines.Line2D(x, y, lw=2., color=color)
            ax2.add_line(line)

            x = np.array([0.3, x.min()])
            y = np.array([starting, y[0]])
            line = lines.Line2D(x, y, lw=2., color=color, alpha=0.4, linestyle='dashed')
            ax2.add_line(line)
            ax2.text(0.5 + length_plot, y[-1], 'F=' + str(F) + ' ', fontsize=20, fontdict={'horizontalalignment': 'left', 'verticalalignment': 'center'})

        for i, label in enumerate(self.ftof):
            lower, upper = label.split('__')
            if '_' in lower:
                lower = lower.split('_')
                lower = float(lower[0]) / float(lower[1])
            else:
                lower = float(lower)
            if '_' in upper:
                upper = upper.split('_')
                upper = float(upper[0]) / float(upper[1])
            else:
                upper = float(upper)

            x = np.array([distances[i], distances[i]]) + 0.5
            lower = energies[np.where(self.F[:self.num_lower]==lower)[0]]
            upper = energies[np.where(self.F[self.num_lower:]==upper)[0] + self.num_lower]
            y = np.array([lower + lower_state_height, upper + upper_state_height]).flatten()
            ax2.arrow(x[0], y[0], x[1] - x[0], y[1] - y[0], fc=arrow_color, ec=arrow_color, length_includes_head=True, head_width=0.02, overhang=0.5, zorder=10)

        ax2.text(0.15, lower_state_height + 0.015, 'J=' + str(Fraction.from_float(self.J[0])), fontsize=20, fontdict={'horizontalalignment': 'center'})
        ax2.text(0.15, upper_state_height + 0.015, 'J=' + str(Fraction.from_float(self.J[-1])), fontsize=20, fontdict={'horizontalalignment': 'center'})
        ax2.text(0.15, np.mean([lower_state_height, upper_state_height]) + 0.015, 'I=' + str(Fraction.from_float(self.I)), fontsize=20, fontdict={'horizontalalignment': 'right'})
        if show:
            plt.show()
        return fig, (ax, ax2)



"""
Implementation of a class for the simultaneous fitting of hyperfine structure spectra.

.. moduleauthor:: Wouter Gins <wouter.gins@kuleuven.be>
.. moduleauthor:: Ruben de Groote <ruben.degroote@kuleuven.be>
"""
import copy

import lmfit as lm
from satlas.models.basemodel import BaseModel
from satlas.utilities import poisson_interval
import matplotlib.pyplot as plt
import numpy as np

__all__ = ['LinkedModel']


class LinkedModel(BaseModel):

    """Links different models for simultaneous fitting."""

    def __init__(self, models):
        """Initializes the class for simultaneous fitting of different models.

        Parameters
        ----------
        models: list of :class:`.BaseModel` children objects
            A list defining the different models."""
        super(LinkedModel, self).__init__()
        self.models = models
        for i, model in enumerate(self.models):
            model._add_prefix('s' + str(i) + '_')
        self._set_params()
        self.shared = []

    def _set_params(self):
        for model in self.models:
            try:
                p.add_many(*model.params.values())
            except:
                p = model.params.copy()
        self.params = p

    def _add_prefix(self, value):
        for model in self.models:
            model._add_prefix(value)
        self._set_params()

    def get_chisquare_mapping(self):
        return np.hstack([f.get_chisquare_mapping() for f in self.models])

    def get_lnprior_mapping(self, params):
        return sum([f.get_lnprior_mapping(f.params) for f in self.models])

    @property
    def shared(self):
        """Contains all parameters which share the same value among all models."""
        return self._shared

    @shared.setter
    def shared(self, value):
        params = self.params.copy()
        self._shared = value
        for name in self._shared:
            selected_list = [p for p in params.keys() if name in p]
            try:
                selected_name = selected_list[0]
                for p in selected_list[1:]:
                    params[p].expr = selected_name
            except IndexError:
                pass
        self.params = params

    @property
    def params(self):
        """Instance of lmfit.Parameters object characterizing the
        shape of the HFS."""
        return self._parameters

    @params.setter
    def params(self, params):
        self._parameters = params.copy()
        for spec in self.models:
            spec.params = self._parameters.copy()
            spec._parameters._prefix = spec._prefix

    def seperate_response(self, x):
        """Generates the response for each subspectrum.

        Parameters
        ----------
        x: list of arrays
            A list equal in length to the number of submodels,
            contains arrays for which the submodels have to be
            evaluated.

        Returns
        -------
        evaluated: ndarray
            The output array, of the same shape as the input
            list of arrays, containing the response values."""
        return np.squeeze([s(X)
                           for s, X in zip(self.models, x)])

    ###############################
    #      PLOTTING ROUTINES      #
    ###############################

    def plot(self, x=None, y=None, yerr=None,
             ax=None, show=True, plot_kws={}, linked=True):
        """Routine that plots the hfs, possibly on top of experimental data.

        Parameters
        ----------
        x: list of arrays
            Experimental x-data. If None, a suitable region around
            the peaks is chosen to plot the hfs.
        y: list of arrays
            Experimental y-data.
        yerr: list of arrays or dict('high': array, 'low': array)
            Experimental errors on y.
        ax: matplotlib axes object
            If provided, plots on these axes.
        show: boolean
            If True, the plot will be shown at the end.
        plot_kws: dictionary
            Dictionary containing the additional keyword arguments for the *plot*
            method of the underlying models. Note that the keyword *ax*
            is passed along correctly.
        linked: boolean, optional
            If True, the x-axes of the generated plots will be linked.

        Returns
        -------
        fig, ax: matplotlib figure and axis
            Figure and axis used for the plotting."""
        if ax is None:
            fig, ax = plt.subplots(len(self.models), 1, sharex=linked)
            height = fig.get_figheight()
            width = fig.get_figwidth()
            fig.set_size_inches(width, len(self.models) * height, forward=True)
        else:
            fig = ax[0].get_figure()
        toReturn = fig, ax

        if x is None:
            x = [None] * len(self.models)
        if y is None:
            y = [None] * len(self.models)
        if yerr is None:
            yerr = [None] * len(self.models)

        selected = int(np.floor(len(self.models)/2 - 1))

        plot_kws_no_xlabel = copy.deepcopy(plot_kws)
        plot_kws_no_xlabel['xlabel'] = ''

        plot_kws_no_ylabel = copy.deepcopy(plot_kws)
        plot_kws_no_ylabel['ylabel'] = ''

        plot_kws_no_xlabel_no_ylabel = copy.deepcopy(plot_kws)
        plot_kws_no_xlabel_no_ylabel['xlabel'] = ''
        plot_kws_no_xlabel_no_ylabel['ylabel'] = ''

        for i, (X, Y, YERR, spec) in enumerate(zip(x, y, yerr,
                                                   self.models)):
            if i == selected:
                try:
                    spec.plot(x=X, y=Y, yerr=[YERR['low'], YERR['high']], ax=ax[i], show=False, plot_kws=plot_kws_no_xlabel)
                except:
                    spec.plot(x=X, y=Y, yerr=YERR, ax=ax[i], show=False, plot_kws=plot_kws_no_xlabel)
            elif i == len(self.models) - 1:
                try:
                    spec.plot(x=X, y=Y, yerr=[YERR['low'], YERR['high']], ax=ax[i], show=False, plot_kws=plot_kws_no_ylabel)
                except:
                    spec.plot(x=X, y=Y, yerr=YERR, ax=ax[i], show=False, plot_kws=plot_kws_no_ylabel)
            else:
                try:
                    spec.plot(x=X, y=Y, yerr=[YERR['low'], YERR['high']], ax=ax[i], show=False, plot_kws=plot_kws_no_xlabel_no_ylabel)
                except:
                    spec.plot(x=X, y=Y, yerr=YERR, ax=ax[i], show=False, plot_kws=plot_kws_no_xlabel_no_ylabel)

        # plt.tight_layout()
        if show:
            plt.show()
        return toReturn

    def plot_spectroscopic(self, x=None, y=None, plot_kws={}, **kwargs):
        """Routine that plots the hfs of all the models, possibly on
        top of experimental data. It assumes that the y data is drawn from
        a Poisson distribution (e.g. counting data).

        Parameters
        ----------
        x: list of arrays
            Experimental x-data. If list of Nones, a suitable region around
            the peaks is chosen to plot the hfs.
        y: list of arrays
            Experimental y-data.
        plot_kws: dictionary
            Dictionary with keys to be passed on to :meth:`.plot`.

        Returns
        -------
        fig, ax: matplotlib figure and axis
            Figure and axis used for the plotting."""

        yerr = []
        if y is not None:
            for Y in y:
                ylow, yhigh = poisson_interval(Y)
                yerr.append({'low': Y - ylow, 'high': yhigh - Y})
        else:
            yerr = None
        return self.plot(x=x, y=y, yerr=yerr, plot_kws=plot_kws, **kwargs)

    ###########################
    #      MAGIC METHODS      #
    ###########################

    def __add__(self, other):
        """Adding another LinkedModel adds the models therein
        to the list of models.

        Returns
        -------
        LinkedModel"""
        if isinstance(other, LinkedModel):
            return_object = LinkedModel(self.models.extend(other.models))
        else:
            return_object = super(LinkedModel, self).__add__(other)
        return return_object

    def __call__(self, x):
        """Pass the seperate frequency arrays to the submodels,
        and return their response values as a list of arrays.

        Parameters
        ----------
        x : list of floats or array_likes
            Frequency in MHz

        Returns
        -------
        list of floats or NumPy arrays
            Response of each spectrum for each seperate value in *x*."""
        return np.hstack([s(X) for s, X in zip(self.models, x)])



"""
Implementation of various functions that ease the work, but do not belong in one of the other modules.

.. moduleauthor:: Wouter Gins <wouter.gins@kuleuven.be>
"""
import numpy as np

__all__ = ['weighted_average',
           'generate_spectrum',
           'poisson_interval',
           'beta',
           'dopplerfactor']

def weighted_average(x, sigma, axis=None):
    r"""Takes the weighted average of an array of values and the associated
    errors. Calculates the scatter and statistical error, and returns
    the greater of these two values.

    Parameters
    ----------
    x: array_like
        Array-like assortment of measured values, is transformed into a
        1D-array.
    sigma: array_like
        Array-like assortment of errors on the measured values, is transformed
        into a 1D-array.

    Returns
    -------
    tuple
        Returns a tuple (weighted_average, uncertainty), with the uncertainty
        being the greater of the uncertainty calculated from the statistical
        uncertainty and the scattering uncertainty.

    Note
    ----
    The formulas used are

    .. math::

        \left\langle x\right\rangle_{weighted} &= \frac{\sum_{i=1}^N \frac{x_i}
                                                                 {\sigma_i^2}}
                                                      {\sum_{i=1}^N \frac{1}
                                                                {\sigma_i^2}}

        \sigma_{stat}^2 &= \frac{1}{\sum_{i=1}^N \frac{1}{\sigma_i^2}}

        \sigma_{scatter}^2 &= \frac{\sum_{i=1}^N \left(\frac{x_i-\left\langle
                                                    x\right\rangle_{weighted}}
                                                      {\sigma_i}\right)^2}
               {\left(N-1\right)\sum_{i=1}^N \frac{1}{\sigma_i^2}}"""
    # x = np.ravel(x)
    # sigma = np.ravel(sigma)
    Xstat = (1 / sigma**2).sum(axis=axis)
    Xm = (x / sigma**2).sum(axis=axis) / Xstat
    # Xscatt = (((x - Xm) / sigma)**2).sum() / ((1 - 1.0 / len(x)) * Xstat)
    Xscatt = (((x - Xm) / sigma)**2).sum(axis=axis) / ((len(x) - 1) * Xstat)
    Xstat = 1 / Xstat
    return Xm, np.maximum.reduce([Xstat, Xscatt], axis=axis) ** 0.5

def generate_spectrum(spectrum, x, number_of_counts, nwalkers=100):
    """Generates a model by random sampling from the provided :class:`.HFSModel`
    and range. The total number of counts for the generated spectrum
    is required.

    Parameters
    ----------
    spectrum: :class:`.HFSModel`
        An instance of class:`.HFSModel`, which gives the probability distribution
        from which the random samples are drawn.
    x: NumPy array
        NumPy array representing the bin centers for the spectrum.
    number_of_counts: int
        Parameter controlling the total number of counts in the spectrum.
    nwalkers: int, optional
        Number of walkers for the random sampling algorithm from emcee.

    Returns
    -------
    y: NumPy array
        Array containing the number of counts corresponding to each value
        in x.
    """
    binsize = x[1] - x[0]  # Need the binsize for accurate lnprob boundaries

    def lnprob(x, left, right):
        if x > right + binsize / 2 or x < left - binsize / 2:
            return -np.inf  # Make sure only to draw from the provided range
        else:
            return np.log(spectrum(x))  # No need to normalize lnprob!
    ndim = 1
    pos = (np.random.rand(nwalkers) * (x.max() - x.min())
           + x.min()).reshape((nwalkers, ndim))
    sampler = mcmc.EnsembleSampler(nwalkers, ndim, lnprob,
                                   args=(x.min(), x.max()))
    # Burn-in
    pos, prob, state = sampler.run_mcmc(pos, 1000)
    sampler.reset()
    # Making sure not to do too much work! Divide requested number of samples
    # by number of walkers, make sure it's a higher integer.
    sampler.run_mcmc(pos, np.ceil(number_of_counts / nwalkers))
    samples = sampler.flatchain[-number_of_counts:]
    # Bin the samples
    bins = x - binsize / 2
    bins = np.append(bins, bins[-1] + binsize)
    y, _ = np.histogram(samples, bins)
    return y

def poisson_interval(data, alpha=0.32):
    """Calculates the confidence interval
    for the mean of a Poisson distribution.

    Parameters
    ----------
    data: array_like
        Data giving the mean of the Poisson distributions.
    alpha: float
        Significance level of interval. Defaults to
        one sigma (0.32).

    Returns
    -------
    low, high: array_like
        Lower and higher limits for the interval."""
    a = alpha
    low, high = (chi2.ppf(a / 2, 2 * data) / 2,
                 chi2.ppf(1 - a / 2, 2 * data + 2) / 2)
    low = np.nan_to_num(low)
    return low, high

def beta(mass, V):
    r"""Calculates the beta-factor for a mass in amu
    and applied voltage in Volt. The formula used is

    .. math::

        \beta = \sqrt{1-\frac{m^2c^4}{\left(mc^2+eV\right)^2}}

    Parameters
    ----------
    mass : float
        Mass in amu.
    V : float
        voltage in volt.

    Returns
    -------
    float
        Relativistic beta-factor.
    """
    c = 299792458.0
    q = 1.60217657 * (10 ** (-19))
    AMU2KG = 1.66053892 * 10 ** (-27)
    mass = mass * AMU2KG
    top = mass ** 2 * c ** 4
    bottom = (mass * c ** 2 + q * V) ** 2
    beta = np.sqrt(1 - top / bottom)
    return beta

def dopplerfactor(mass, V):
    r"""Calculates the Doppler shift of the laser frequency for a
    given mass in amu and voltage in V. Transforms from the lab frame
    to the particle frame. The formula used is

    .. math::

        doppler = \sqrt{\frac{1-\beta}{1+\beta}}

    To invert, divide instead of multiply with
    this factor.

    Parameters
    ----------
    mass : float
        Mass in amu.
    V : float
        Voltage in volt.

    Returns
    -------
    float
        Doppler factor.
    """
    betaFactor = beta(mass, V)
    dopplerFactor = np.sqrt((1.0 - betaFactor) / (1.0 + betaFactor))
    return dopplerFactor

def extract_result(filename, filter=None, bins=None, selection=(0, 100)):
    with h5py.File(filename, 'r') as store:
        columns = store['data'].attrs['format']
        columns = [f.decode('utf-8') for f in columns]
        if filter is not None:
            filter = [c for f in filter for c in columns if f in c]
        else:
            filter = columns
        metadata = {}
        if not isinstance(bins, list):
            bins = [bins for _ in filter]
        dataset_length = store['data'].shape[0]
        first, last = int(np.floor(dataset_length/100*selection[0])), int(np.ceil(dataset_length/100*selection[1]))
        for i, val in enumerate(filter):
            bin_index = i
            i = columns.index(val)
            x = store['data'][first:last, i]
            if bins[bin_index] is None:
                width = 3.5*np.std(x)/x.size**(1/3)
                bins[bin_index] = np.arange(x.min(), x.max()+width, width)
            try:
                n, b, = np.histogram(x, int(bins[bin_index]))
            except TypeError:
                bins[bin_index] = 50
                n, b, = np.histogram(x, int(bins[bin_index]))
            center = n.argmax()
            q50 = (b[center] + b[center+1])/2

            q = [16.0, 84.0]
            q16, q84 = np.percentile(x, q)
            metadata[val] = {'50': q50, 'up': q84-q50, 'down': q50-q16}
    return metadata



"""
Implementation of a class for the analysis of hyperfine structure spectra with isomeric presence.

.. moduleauthor:: Wouter Gins <wouter.gins@kuleuven.be>
.. moduleauthor:: Ruben de Groote <ruben.degroote@kuleuven.be>
"""
import copy
from satlas.models.basemodel import BaseModel
from satlas.utilities import poisson_interval
import matplotlib.pyplot as plt
import numpy as np


__all__ = ['SumModel']


class SumModel(BaseModel):

    """Create a model that sums all the underlying models for a single input variable."""

    def __init__(self, models):
        """Initializes the HFS by providing a list of :class:`.HFSModel`
        objects.

        Parameters
        ----------
        models: list of :class:`.HFSModel` instances
            A list containing the models."""
        super(SumModel, self).__init__()
        self.models = models
        for i, model in enumerate(self.models):
            model._add_prefix('s' + str(i) + '_')
        self._set_params()
        self.shared = []

    def _set_params(self):
        for model in self.models:
            try:
                p.add_many(*model.params.values())
            except:
                p = model.params.copy()
        self.params = p

    def _add_prefix(self, value):
        for model in self.models:
            model._add_prefix(value)
        self._set_params()

    def get_chisquare_mapping(self):
        return np.hstack([f.get_chisquare_mapping() for f in self.models])

    def get_lnprior_mapping(self, params):
        return sum([f.get_lnprior_mapping(params) for f in self.models])

    @property
    def shared(self):
        """Contains all parameters which share the same value among all models."""
        return self._shared

    @shared.setter
    def shared(self, value):
        params = self.params.copy()
        self._shared = value
        for name in self._shared:
            selected_list = [p for p in params.keys() if name in p]
            try:
                selected_name = selected_list[0]
                for p in selected_list[1:]:
                    params[p].expr = selected_name
            except IndexError:
                pass
        self.params = params

    @property
    def params(self):
        """Instance of lmfit.Parameters object characterizing the
        shape of the HFS."""
        return self._parameters

    @params.setter
    def params(self, params):
        self._parameters = params.copy()
        for spec in self.models:
            spec.params = self._parameters.copy()

    def seperate_response(self, x, background=False):
        """Get the response for each seperate spectrum for the values *x*,
        without background.

        Parameters
        ----------
        x : float or array_like
            Frequency in MHz.

        Other parameters
        ----------------
        background: boolean
            If True, each spectrum has the same background. If False,
            the background of each spectrum is assumed to be 0.

        Returns
        -------
        list of floats or NumPy arrays
            Seperate responses of models to the input *x*."""
        background_vals = [np.polyval([s.params[par_name].value for par_name in s.params if par_name.startswith('Background')], x) for s in self.models]
        return [s(x) - b * (1-background) for s, b in zip(self.models, background_vals)]

    ###############################
    #      PLOTTING ROUTINES      #
    ###############################

    def plot(self, x=None, y=None, yerr=None, ax=None, plot_kws={}, plot_seperate=True, show=True, legend=None, data_legend=None, xlabel='Frequency (MHz)', ylabel='Counts'):
        """Routine that plots the hfs of all the models,
        possibly on top of experimental data.

        Parameters
        ----------
        x: list of arrays
            Experimental x-data. If list of Nones, a suitable region around
            the peaks is chosen to plot the hfs.
        y: list of arrays
            Experimental y-data.
        yerr: list of arrays
            Experimental errors on y.
        plot_seperate: boolean, optional
            Controls if the underlying models are drawn as well, or only
            the sum. Defaults to False.
        no_of_points: int
            Number of points to use for the plot of the hfs if
            experimental data is given.
        ax: matplotlib axes object
            If provided, plots on this axis.
        show: boolean
            If True, the plot will be shown at the end.
        legend: string, optional
            If given, an entry in the legend will be made for the spectrum.
        data_legend: string, optional
            If given, an entry in the legend will be made for the experimental
            data.
        xlabel: string, optional
            If given, sets the xlabel to this string. Defaults to 'Frequency (MHz)'.
        ylabel: string, optional
            If given, sets the ylabel to this string. Defaults to 'Counts'.
        indicate: boolean, optional
            If set to True, dashed lines are drawn to indicate the location of the
            transitions, and the labels are attached. Defaults to False.
        model: boolean, optional
            If given, the region around the fitted line will be shaded, with
            the luminosity indicating the pmf of the Poisson
            distribution characterized by the value of the fit. Note that
            the argument *yerr* is ignored if *model* is True.
        normalized: Boolean
            If True, the data and fit are plotted normalized such that the highest
            data point is one.
        distance: float, optional
            Controls how many FWHM deviations are used to generate the plot.
            Defaults to 4.

        Returns
        -------
        fig, ax: matplotlib figure and axes"""
        if ax is None:
            fig, ax = plt.subplots(1, 1)
        else:
            fig = ax.get_figure()
        toReturn = fig, ax

        if x is not None and y is not None:
            try:
                ax.errorbar(x, y, yerr=[y - yerr['low'], yerr['high'] - y], fmt='o', label=data_legend)
            except:
                ax.errorbar(x, y, yerr=yerr, fmt='o', label=data_legend)

        plot_kws['background'] = False
        plot_copy = copy.deepcopy(plot_kws)
        plot_copy['model'] = False
        x_points = np.array([])
        line_counter = 1
        for m in self.models:
            plot_copy['legend'] = 'I=' + str(m.I)
            try:
                color = ax.lines[-1].get_color()
            except IndexError:
                color = next(ax._get_lines.prop_cycler)['color']
            m.plot(x=x, y=y, yerr=yerr, show=False, ax=ax, plot_kws=plot_copy)
            # plot_kws['indicate'] = False
            x_points = np.append(x_points, ax.lines[-1].get_xdata())
            if not plot_seperate:
                ax.lines.pop(-1)
            if x is not None:
                ax.lines.pop(-1 - plot_seperate)
            while not next(ax._get_lines.prop_cycler)['color'] == color:
                pass
            if plot_seperate:
                c = next(ax._get_lines.prop_cycler)['color']
                for l in ax.lines[line_counter:]:
                    l.set_color(c)
                while not next(ax._get_lines.prop_cycler)['color'] == c:
                    pass
            line_counter = len(ax.lines)
        x = np.sort(x_points)
        model = plot_kws.pop('model', False)
        if model:
            colormap = plot_kws.pop('colormap', 'bone_r',)
            min_loc = [s.locations.min() for s in self.models]
            max_loc = [s.locations.max() for s in self.models]
            range = (min(min_loc), max(max_loc))
            from scipy import optimize
            max_counts = np.ceil(-optimize.brute(lambda x: -self(x), (range,), full_output=True, Ns=1000, finish=optimize.fmin)[1])
            min_counts = [self.params[par_name].value for par_name in self.params if par_name.endswith('Background0')][0]
            min_counts = np.floor(max(0, min_counts - 3 * min_counts ** 0.5))
            y = np.arange(min_counts, max_counts + 3 * max_counts ** 0.5 + 1)
            X, Y = np.meshgrid(x, y)
            from scipy import stats
            z = stats.poisson(self(X)).pmf(Y)

            z = z / z.sum(axis=0)
            ax.imshow(z, extent=(x.min(), x.max(), y.min(), y.max()), cmap=plt.get_cmap(colormap))
            line, = ax.plot(x, self(x), label=legend, lw=0.5)
        else:
            ax.plot(x, self(x))
        ax.legend(loc=0)

        # ax.set_xlabel(xlabel)
        # ax.set_ylabel(ylabel)

        if show:
            plt.show()
        return toReturn

    def plot_spectroscopic(self, x=None, y=None, plot_kws={}, **kwargs):
        """Routine that plots the hfs of all the models, possibly on
        top of experimental data. It assumes that the y data is drawn from
        a Poisson distribution (e.g. counting data).

        Parameters
        ----------
        x: list of arrays
            Experimental x-data. If list of Nones, a suitable region around
            the peaks is chosen to plot the hfs.
        y: list of arrays
            Experimental y-data.
        yerr: list of arrays
            Experimental errors on y.
        no_of_points: int
            Number of points to use for the plot of the hfs.
        ax: matplotlib axes object
            If provided, plots on this axis
        show: Boolean
            if True, the plot will be shown at the end.

        Returns
        -------
        fig, ax: matplotlib figure and axes"""

        if y is not None:
            ylow, yhigh = poisson_interval(y)
            yerr = {'low': ylow, 'high': yhigh}
        else:
            yerr = None
        return self.plot(x=x, y=y, yerr=yerr, plot_kws=plot_kws, **kwargs)

    def __add__(self, other):
        """Adding an SumModel results in a new SumModel
        with the new spectrum added.

        Returns
        -------
        SumModel"""
        if isinstance(other, SumModel):
            models = self.models + other.models
            return SumModel(models)
        else:
            try:
                return other.__add__(self)
            except:
                raise TypeError('unsupported operand type(s)')

    def __call__(self, x):
        """Get the response for frequency *x* (in MHz) of the spectrum.

        Parameters
        ----------
        x : float or array_like
            Frequency in MHz

        Returns
        -------
        float or NumPy array
            Response of the spectrum for each value of *x*."""
        return np.sum([s(x) for s in self.models], axis=0)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
An affine invariant Markov chain Monte Carlo (MCMC) sampler.

Goodman & Weare, Ensemble Samplers With Affine Invariance
   Comm. App. Math. Comp. Sci., Vol. 5 (2010), No. 1, 65–80

"""

from __future__ import (division, print_function, absolute_import,
                        unicode_literals)

__all__ = ["EnsembleSampler"]

import numpy as np

from . import autocorr
from .sampler import Sampler
from .interruptible_pool import InterruptiblePool


class EnsembleSampler(Sampler):
    """
    A generalized Ensemble sampler that uses 2 ensembles for parallelization.
    The ``__init__`` function will raise an ``AssertionError`` if
    ``k < 2 * dim`` (and you haven't set the ``live_dangerously`` parameter)
    or if ``k`` is odd.

    **Warning**: The :attr:`chain` member of this object has the shape:
    ``(nwalkers, nlinks, dim)`` where ``nlinks`` is the number of steps
    taken by the chain and ``k`` is the number of walkers.  Use the
    :attr:`flatchain` property to get the chain flattened to
    ``(nlinks, dim)``. For users of pre-1.0 versions, this shape is
    different so be careful!

    :param nwalkers:
        The number of Goodman & Weare "walkers".

    :param dim:
        Number of dimensions in the parameter space.

    :param lnpostfn:
        A function that takes a vector in the parameter space as input and
        returns the natural logarithm of the posterior probability for that
        position.

    :param a: (optional)
        The proposal scale parameter. (default: ``2.0``)

    :param args: (optional)
        A list of extra positional arguments for ``lnpostfn``. ``lnpostfn``
        will be called with the sequence ``lnpostfn(p, *args, **kwargs)``.

    :param kwargs: (optional)
        A list of extra keyword arguments for ``lnpostfn``. ``lnpostfn``
        will be called with the sequence ``lnpostfn(p, *args, **kwargs)``.

    :param postargs: (optional)
        Alias of ``args`` for backwards compatibility.

    :param threads: (optional)
        The number of threads to use for parallelization. If ``threads == 1``,
        then the ``multiprocessing`` module is not used but if
        ``threads > 1``, then a ``Pool`` object is created and calls to
        ``lnpostfn`` are run in parallel.

    :param pool: (optional)
        An alternative method of using the parallelized algorithm. If
        provided, the value of ``threads`` is ignored and the
        object provided by ``pool`` is used for all parallelization. It
        can be any object with a ``map`` method that follows the same
        calling sequence as the built-in ``map`` function.

    :param runtime_sortingfn: (optional)
        A function implementing custom runtime load-balancing. See
        :ref:`loadbalance` for more information.

    """
    def __init__(self, nwalkers, dim, lnpostfn, a=2.0, args=[], kwargs={},
                 postargs=None, threads=1, pool=None, live_dangerously=False,
                 runtime_sortingfn=None):
        self.k = nwalkers
        self.a = a
        self.threads = threads
        self.pool = pool
        self.runtime_sortingfn = runtime_sortingfn

        if postargs is not None:
            args = postargs
        super(EnsembleSampler, self).__init__(dim, lnpostfn, args=args,
                                              kwargs=kwargs)

        # Do a little bit of _magic_ to make the likelihood call with
        # ``args`` and ``kwargs`` pickleable.
        self.lnprobfn = _function_wrapper(self.lnprobfn, self.args,
                                          self.kwargs)

        assert self.k % 2 == 0, "The number of walkers must be even."
        if not live_dangerously:
            assert self.k >= 2 * self.dim, (
                "The number of walkers needs to be more than twice the "
                "dimension of your parameter space... unless you're "
                "crazy!")

        if self.threads > 1 and self.pool is None:
            self.pool = InterruptiblePool(self.threads)

    def clear_blobs(self):
        """
        Clear the ``blobs`` list.

        """
        self._blobs = []

    def reset(self):
        """
        Clear the ``chain`` and ``lnprobability`` array. Also reset the
        bookkeeping parameters.

        """
        super(EnsembleSampler, self).reset()
        self.naccepted = np.zeros(self.k)
        self._chain = np.empty((self.k, 0, self.dim))
        self._lnprob = np.empty((self.k, 0))

        # Initialize list for storing optional metadata blobs.
        self.clear_blobs()

    def sample(self, p0, lnprob0=None, rstate0=None, blobs0=None,
               iterations=1, thin=1, storechain=True, mh_proposal=None):
        """
        Advance the chain ``iterations`` steps as a generator.

        :param p0:
            A list of the initial positions of the walkers in the
            parameter space. It should have the shape ``(nwalkers, dim)``.

        :param lnprob0: (optional)
            The list of log posterior probabilities for the walkers at
            positions given by ``p0``. If ``lnprob is None``, the initial
            values are calculated. It should have the shape ``(k, dim)``.

        :param rstate0: (optional)
            The state of the random number generator.
            See the :attr:`Sampler.random_state` property for details.

        :param iterations: (optional)
            The number of steps to run.

        :param thin: (optional)
            If you only want to store and yield every ``thin`` samples in the
            chain, set thin to an integer greater than 1.

        :param storechain: (optional)
            By default, the sampler stores (in memory) the positions and
            log-probabilities of the samples in the chain. If you are
            using another method to store the samples to a file or if you
            don't need to analyse the samples after the fact (for burn-in
            for example) set ``storechain`` to ``False``.

        :param mh_proposal: (optional)
            A function that returns a list of positions for ``nwalkers``
            walkers given a current list of positions of the same size. See
            :class:`utils.MH_proposal_axisaligned` for an example.

        At each iteration, this generator yields:

        * ``pos`` - A list of the current positions of the walkers in the
          parameter space. The shape of this object will be
          ``(nwalkers, dim)``.

        * ``lnprob`` - The list of log posterior probabilities for the
          walkers at positions given by ``pos`` . The shape of this object
          is ``(nwalkers, dim)``.

        * ``rstate`` - The current state of the random number generator.

        * ``blobs`` - (optional) The metadata "blobs" associated with the
          current position. The value is only returned if ``lnpostfn``
          returns blobs too.

        """
        # Try to set the initial value of the random number generator. This
        # fails silently if it doesn't work but that's what we want because
        # we'll just interpret any garbage as letting the generator stay in
        # it's current state.
        self.random_state = rstate0

        p = np.array(p0)
        halfk = int(self.k / 2)

        # If the initial log-probabilities were not provided, calculate them
        # now.
        lnprob = lnprob0
        blobs = blobs0
        if lnprob is None:
            lnprob, blobs = self._get_lnprob(p)

        # Check to make sure that the probability function didn't return
        # ``np.nan``.
        if np.any(np.isnan(lnprob)):
            raise ValueError("The initial lnprob was NaN.")

        # Store the initial size of the stored chain.
        i0 = self._chain.shape[1]

        # Here, we resize chain in advance for performance. This actually
        # makes a pretty big difference.
        if storechain:
            N = int(iterations / thin)
            self._chain = np.concatenate((self._chain,
                                          np.zeros((self.k, N, self.dim))),
                                         axis=1)
            self._lnprob = np.concatenate((self._lnprob,
                                           np.zeros((self.k, N))), axis=1)

        for i in range(int(iterations)):
            self.iterations += 1

            # If we were passed a Metropolis-Hastings proposal
            # function, use it.
            if mh_proposal is not None:
                # Draw proposed positions & evaluate lnprob there
                q = mh_proposal(p)
                newlnp, blob = self._get_lnprob(q)

                # Accept if newlnp is better; and ...
                acc = (newlnp > lnprob)

                # ... sometimes accept for steps that got worse
                worse = np.flatnonzero(~acc)
                acc[worse] = ((newlnp[worse] - lnprob[worse]) >
                              np.log(self._random.rand(len(worse))))
                del worse

                # Update the accepted walkers.
                lnprob[acc] = newlnp[acc]
                p[acc] = q[acc]
                self.naccepted[acc] += 1

                if blob is not None:
                    assert blobs is not None, (
                        "If you start sampling with a given lnprob, you also "
                        "need to provide the current list of blobs at that "
                        "position.")
                    ind = np.arange(self.k)[acc]
                    for j in ind:
                        blobs[j] = blob[j]

            else:
                # Loop over the two ensembles, calculating the proposed
                # positions.

                # Slices for the first and second halves
                first, second = slice(halfk), slice(halfk, self.k)
                for S0, S1 in [(first, second), (second, first)]:
                    q, newlnp, acc, blob = self._propose_stretch(p[S0], p[S1],
                                                                 lnprob[S0])
                    if np.any(acc):
                        # Update the positions, log probabilities and
                        # acceptance counts.
                        lnprob[S0][acc] = newlnp[acc]
                        p[S0][acc] = q[acc]
                        self.naccepted[S0][acc] += 1

                        if blob is not None:
                            assert blobs is not None, (
                                "If you start sampling with a given lnprob, "
                                "you also need to provide the current list of "
                                "blobs at that position.")
                            ind = np.arange(len(acc))[acc]
                            indfull = np.arange(self.k)[S0][acc]
                            for j in range(len(ind)):
                                blobs[indfull[j]] = blob[ind[j]]

            if storechain and i % thin == 0:
                ind = i0 + int(i / thin)
                self._chain[:, ind, :] = p
                self._lnprob[:, ind] = lnprob
                if blobs is not None:
                    self._blobs.append(list(blobs))

            # Yield the result as an iterator so that the user can do all
            # sorts of fun stuff with the results so far.
            if blobs is not None:
                # This is a bit of a hack to keep things backwards compatible.
                yield p, lnprob, self.random_state, blobs
            else:
                yield p, lnprob, self.random_state

    def _propose_stretch(self, p0, p1, lnprob0):
        """
        Propose a new position for one sub-ensemble given the positions of
        another.

        :param p0:
            The positions from which to jump.

        :param p1:
            The positions of the other ensemble.

        :param lnprob0:
            The log-probabilities at ``p0``.

        This method returns:

        * ``q`` - The new proposed positions for the walkers in ``ensemble``.

        * ``newlnprob`` - The vector of log-probabilities at the positions
          given by ``q``.

        * ``accept`` - A vector of type ``bool`` indicating whether or not
          the proposed position for each walker should be accepted.

        * ``blob`` - The new meta data blobs or ``None`` if nothing was
          returned by ``lnprobfn``.

        """
        s = np.atleast_2d(p0)
        Ns = len(s)
        c = np.atleast_2d(p1)
        Nc = len(c)

        # Generate the vectors of random numbers that will produce the
        # proposal.
        zz = ((self.a - 1.) * self._random.rand(Ns) + 1) ** 2. / self.a
        rint = self._random.randint(Nc, size=(Ns,))

        # Calculate the proposed positions and the log-probability there.
        q = c[rint] - zz[:, np.newaxis] * (c[rint] - s)
        newlnprob, blob = self._get_lnprob(q)

        # Decide whether or not the proposals should be accepted.
        lnpdiff = (self.dim - 1.) * np.log(zz) + newlnprob - lnprob0
        accept = (lnpdiff > np.log(self._random.rand(len(lnpdiff))))

        return q, newlnprob, accept, blob

    def _get_lnprob(self, pos=None):
        """
        Calculate the vector of log-probability for the walkers.

        :param pos: (optional)
            The position vector in parameter space where the probability
            should be calculated. This defaults to the current position
            unless a different one is provided.

        This method returns:

        * ``lnprob`` - A vector of log-probabilities with one entry for each
          walker in this sub-ensemble.

        * ``blob`` - The list of meta data returned by the ``lnpostfn`` at
          this position or ``None`` if nothing was returned.

        """
        if pos is None:
            p = self.pos
        else:
            p = pos

        # Check that the parameters are in physical ranges.
        if np.any(np.isinf(p)):
            raise ValueError("At least one parameter value was infinite.")
        if np.any(np.isnan(p)):
            raise ValueError("At least one parameter value was NaN.")

        # If the `pool` property of the sampler has been set (i.e. we want
        # to use `multiprocessing`), use the `pool`'s map method. Otherwise,
        # just use the built-in `map` function.
        if self.pool is not None:
            M = self.pool.map
        else:
            M = map

        # sort the tasks according to (user-defined) some runtime guess
        if self.runtime_sortingfn is not None:
            p, idx = self.runtime_sortingfn(p)

        # Run the log-probability calculations (optionally in parallel).
        results = list(M(self.lnprobfn, [p[i] for i in range(len(p))]))

        try:
            lnprob = np.array([float(l[0]) for l in results])
            blob = [l[1] for l in results]
        except (IndexError, TypeError):
            lnprob = np.array([float(l) for l in results])
            blob = None

        # sort it back according to the original order - get the same
        # chain irrespective of the runtime sorting fn
        if self.runtime_sortingfn is not None:
            orig_idx = np.argsort(idx)
            lnprob = lnprob[orig_idx]
            p = [p[i] for i in orig_idx]
            if blob is not None:
                blob = [blob[i] for i in orig_idx]

        # Check for lnprob returning NaN.
        if np.any(np.isnan(lnprob)):
            # Print some debugging stuff.
            print("NaN value of lnprob for parameters: ")
            for pars in p[np.isnan(lnprob)]:
                print(pars)

            # Finally raise exception.
            raise ValueError("lnprob returned NaN.")

        return lnprob, blob

    @property
    def blobs(self):
        """
        Get the list of "blobs" produced by sampling. The result is a list
        (of length ``iterations``) of ``list`` s (of length ``nwalkers``) of
        arbitrary objects. **Note**: this will actually be an empty list if
        your ``lnpostfn`` doesn't return any metadata.

        """
        return self._blobs

    @property
    def chain(self):
        """
        A pointer to the Markov chain itself. The shape of this array is
        ``(k, iterations, dim)``.

        """
        return super(EnsembleSampler, self).chain

    @property
    def flatchain(self):
        """
        A shortcut for accessing chain flattened along the zeroth (walker)
        axis.

        """
        s = self.chain.shape
        return self.chain.reshape(s[0] * s[1], s[2])

    @property
    def lnprobability(self):
        """
        A pointer to the matrix of the value of ``lnprobfn`` produced at each
        step for each walker. The shape is ``(k, iterations)``.

        """
        return super(EnsembleSampler, self).lnprobability

    @property
    def flatlnprobability(self):
        """
        A shortcut to return the equivalent of ``lnprobability`` but aligned
        to ``flatchain`` rather than ``chain``.

        """
        return super(EnsembleSampler, self).lnprobability.flatten()

    @property
    def acceptance_fraction(self):
        """
        An array (length: ``k``) of the fraction of steps accepted for each
        walker.

        """
        return super(EnsembleSampler, self).acceptance_fraction

    @property
    def acor(self):
        """
        An estimate of the autocorrelation time for each parameter (length:
        ``dim``).

        """
        return self.get_autocorr_time()

    def get_autocorr_time(self, window=50, fast=False):
        """
        Compute an estimate of the autocorrelation time for each parameter
        (length: ``dim``).

        :param window: (optional)
            The size of the windowing function. This is equivalent to the
            maximum number of lags to use. (default: 50)

        """
        return autocorr.integrated_time(np.mean(self.chain, axis=0), axis=0,
                                        window=window, fast=fast)


class _function_wrapper(object):
    """
    This is a hack to make the likelihood function pickleable when ``args``
    or ``kwargs`` are also included.

    """
    def __init__(self, f, args, kwargs):
        self.f = f
        self.args = args
        self.kwargs = kwargs

    def __call__(self, x):
        try:
            return self.f(x, *self.args, **self.kwargs)
        except:
            import traceback
            print("emcee: Exception while calling your likelihood function:")
            print("  params:", x)
            print("  args:", self.args)
            print("  kwargs:", self.kwargs)
            print("  exception:")
            traceback.print_exc()
            raise



"""
Implementation of a class for the analysis of linear data.

.. moduleauthor:: Wouter Gins <wouter.gins@kuleuven.be>
"""

import lmfit as lm
from satlas.models.basemodel import BaseModel, SATLASParameters
import numpy as np

__all__ = ['PolynomialModel', 'MiscModel']


class PolynomialModel(BaseModel):

    r"""Constructs a polynomial response."""

    def __init__(self, args):
        """:class:`.PolynomialModel` creates a general polynomial
        of the order given by *len(args)-1*. The given coefficients
        are ordered lowest to highest order.

        Parameters
        ----------
        args: iterable of values
            Iterable containing all the values for the
            coefficients. Polynomial order is determined
            by the length. args[0] is the coefficient
            of order 0, etc..."""
        super(PolynomialModel, self).__init__()
        self._populate_params(args)

    @property
    def params(self):
        return self._parameters

    @params.setter
    def params(self, params):
        p = params.copy()
        p._prefix = self._prefix
        self._parameters = self._check_variation(p)

    ####################################
    #      INITIALIZATION METHODS      #
    ####################################

    def _populate_params(self, args):
        # Prepares the params attribute with the initial values
        par = SATLASParameters()
        for i, val in reversed(list(enumerate(args))):
            par.add('Order' + str(i) + 'Coeff', value=val, vary=True)
        self.params = self._check_variation(par)

        self.names = [p for p in self.params.keys()]

    ###########################
    #      MAGIC METHODS      #
    ###########################

    def __call__(self, x):
        return np.polyval([self.params[p].value for p in self.names], x)

class MiscModel(BaseModel):

    r"""Constructs a response from a supplied function.
    Call signature is

    def func(x, par):

        a = par[0]

        b = par[1]

        ...

        return y"""

    def __init__(self, func, args, name_list=None):
        """The :class:`.MiscModel` takes a supplied function *func* and list of starting
        argument parameters *args* to contruct an object that responds with the
        given function for the parameter values. A list of names can also
        be supplied to customize the parameter names.

        Parameters
        ----------
        func: callable
            A callable function with call signature *func(x, args)*.
        args: list of values
            List of starting values for the parameters. The number of parameters is based
            on the length of the list of arguments.
        name_list: list of strings, optional
            List of names to be supplied to the parameters. The order of the names
            and the order of the parameters is the same, so *name_list[0]* corresponds
            to *args[0]*."""
        super(MiscModel, self).__init__()
        self.func = func
        self.name_list = name_list
        self._populate_params(*args)

    @property
    def params(self):
        return self._parameters

    @params.setter
    def params(self, params):
        p = params.copy()
        p._prefix = self._prefix
        self._parameters = self._check_variation(p)

    ####################################
    #      INITIALIZATION METHODS      #
    ####################################

    def _populate_params(self, *args):
        # Prepares the params attribute with the initial values
        par = SATLASParameters()
        names = []
        if self.name_list is None:
            for i, val in enumerate(args):
                par.add('Param' + str(i + 1), value=val, vary=True)
                names.append('Param' + str(i + 1))
        else:
            for name, val in zip(self.name_list, args):
                par.add(name, value=val, vary=True)
                names.append(name)

        self.params = self._check_variation(par)
        self.names = names

    ###########################
    #      MAGIC METHODS      #
    ###########################

    def __call__(self, x):
        return self.func(x, [self.params[p].value for p in self.names])



"""
Customisable progressbar decorator for iterators.
Includes a default (x)range iterator printing to stderr.

Usage:
  >>> from tqdm import trange[, tqdm]
  >>> for i in trange(10): #same as: for i in tqdm(xrange(10))
  ...     ...
"""
# future division is important to divide integers and get as
# a result precise floating numbers (instead of truncated int)
from __future__ import division, absolute_import
# import compatibility functions and utilities
from ._utils import _supports_unicode, _environ_cols_wrapper, _range, _unich, \
    _term_move_up, _unicode
import sys
from time import time


__author__ = {"github.com/": ["noamraph", "obiwanus", "kmike", "hadim",
                              "casperdcl", "lrq3000"]}
__all__ = ['tqdm', 'trange']


class tqdm(object):
    """
    Decorate an iterable object, returning an iterator which acts exactly
    like the original iterable, but prints a dynamically updating
    progressbar every time a value is requested.
    """
    @staticmethod
    def format_sizeof(num, suffix=''):
        """
        Formats a number (greater than unity) with SI Order of Magnitude
        prefixes.

        Parameters
        ----------
        num  : float
            Number ( >= 1) to format.
        suffix  : str, optional
            Post-postfix [default: ''].

        Returns
        -------
        out  : str
            Number with Order of Magnitude SI unit postfix.
        """
        for unit in ['', 'K', 'M', 'G', 'T', 'P', 'E', 'Z']:
            if abs(num) < 999.95:
                if abs(num) < 99.95:
                    if abs(num) < 9.995:
                        return '{0:1.2f}'.format(num) + unit + suffix
                    return '{0:2.1f}'.format(num) + unit + suffix
                return '{0:3.0f}'.format(num) + unit + suffix
            num /= 1000.0
        return '{0:3.1f}Y'.format(num) + suffix

    @staticmethod
    def format_interval(t):
        """
        Formats a number of seconds as a clock time, [H:]MM:SS

        Parameters
        ----------
        t  : int
            Number of seconds.
        Returns
        -------
        out  : str
            [H:]MM:SS
        """
        mins, s = divmod(int(t), 60)
        h, m = divmod(mins, 60)
        if h:
            return '{0:d}:{1:02d}:{2:02d}'.format(h, m, s)
        else:
            return '{0:02d}:{1:02d}'.format(m, s)

    @staticmethod
    def status_printer(file):
        """
        Manage the printing and in-place updating of a line of characters.
        Note that if the string is longer than a line, then in-place
        updating may not work (it will print a new line at each refresh).
        """
        fp = file
        if not getattr(fp, 'flush', False):  # pragma: no cover
            fp.flush = lambda: None

        last_printed_len = [0]  # closure over mutable variable (fast)

        def print_status(s):
            len_s = len(s)
            fp.write('\r' + s + (' ' * max(last_printed_len[0] - len_s, 0)))
            fp.flush()
            last_printed_len[0] = len_s
        return print_status

    @staticmethod
    def format_meter(n, total, elapsed, ncols=None, prefix='',
                     ascii=False, unit='it', unit_scale=False, rate=None,
                     bar_format=None):
        """
        Return a string-based progress bar given some parameters

        Parameters
        ----------
        n  : int
            Number of finished iterations.
        total  : int
            The expected total number of iterations. If meaningless (), only
            basic progress statistics are displayed (no ETA).
        elapsed  : float
            Number of seconds passed since start.
        ncols  : int, optional
            The width of the entire output message. If specified,
            dynamically resizes the progress meter to stay within this bound
            [default: None]. The fallback meter width is 10 for the progress
            bar + no limit for the iterations counter and statistics. If 0,
            will not print any meter (only stats).
        prefix  : str, optional
            Prefix message (included in total width) [default: ''].
        ascii  : bool, optional
            If not set, use unicode (smooth blocks) to fill the meter
            [default: False]. The fallback is to use ASCII characters
            (1-9 #).
        unit  : str, optional
            The iteration unit [default: 'it'].
        unit_scale  : bool, optional
            If set, the number of iterations will printed with an
            appropriate SI metric prefix (K = 10^3, M = 10^6, etc.)
            [default: False].
        rate  : float, optional
            Manual override for iteration rate.
            If [default: None], uses n/elapsed.
        bar_format  : str, optional
            Specify a custom bar string formatting. May impact performance.
            [default: '{l_bar}{bar}{r_bar}'], where l_bar is
            '{desc}{percentage:3.0f}%|' and r_bar is
            '| {n_fmt}/{total_fmt} [{elapsed_str}<{remaining_str}, {rate_fmt}]'.
            Possible vars: bar, n, n_fmt, total, total_fmt, percentage,
            rate, rate_fmt, elapsed, remaining, l_bar, r_bar, desc.

        Returns
        -------
        out  : Formatted meter and stats, ready to display.
        """

        # sanity check: total
        if total and n > total:
            total = None

        format_interval = tqdm.format_interval
        elapsed_str = format_interval(elapsed)

        # if unspecified, attempt to use rate = average speed
        # (we allow manual override since predicting time is an arcane art)
        if rate is None and elapsed:
            rate = n / elapsed
        inv_rate = 1 / rate if (rate and (rate < 1)) else None
        format_sizeof = tqdm.format_sizeof
        rate_fmt = ((format_sizeof(inv_rate if inv_rate else rate)
                    if unit_scale else
                    '{0:5.2f}'.format(inv_rate if inv_rate else rate))
                    if elapsed else '?') \
            + ('s' if inv_rate else unit) + '/' + (unit if inv_rate else 's')

        if unit_scale:
            n_fmt = format_sizeof(n)
            total_fmt = format_sizeof(total) if total else None
        else:
            n_fmt = str(n)
            total_fmt = str(total)

        # total is known: we can predict some stats
        if total:
            # fractional and percentage progress
            frac = n / total
            percentage = frac * 100

            remaining_str = format_interval((total - n) / rate) \
                if rate else '?'

            # format the stats displayed to the left and right sides of the bar
            l_bar = (prefix if prefix else '') + \
                '{0:3.0f}%|'.format(percentage)
            r_bar = '| {0}/{1} [{2}<{3}, {4}]'.format(
                    n_fmt, total_fmt, elapsed_str, remaining_str, rate_fmt)

            if ncols == 0:
                return l_bar[:-1] + r_bar[1:]

            # space available for bar's display
            N_BARS = max(1, ncols - len(l_bar) - len(r_bar)) if ncols \
                else 10

            # format bar depending on availability of unicode/ascii chars
            if ascii:
                bar_length, frac_bar_length = divmod(
                    int(frac * N_BARS * 10), 10)

                bar = '#' * bar_length
                frac_bar = chr(48 + frac_bar_length) if frac_bar_length \
                    else ' '

            else:
                bar_length, frac_bar_length = divmod(int(frac * N_BARS * 8), 8)

                bar = _unich(0x2588) * bar_length
                frac_bar = _unich(0x2590 - frac_bar_length) \
                    if frac_bar_length else ' '

            # whitespace padding
            if bar_length < N_BARS:
                full_bar = bar + frac_bar + \
                    ' ' * max(N_BARS - bar_length - 1, 0)
            else:
                full_bar = bar + \
                    ' ' * max(N_BARS - bar_length, 0)

            if bar_format is None:
                # Default bar format = fast display
                return l_bar + full_bar + r_bar
            else:
                # Custom bar formatting
                # Populate a dict with all available progress indicators
                bar_args = {'bar': full_bar,
                            'n': n,
                            'n_fmt': n_fmt,
                            'total': total,
                            'total_fmt': total_fmt,
                            'percentage': percentage,
                            'rate': rate,
                            'rate_fmt': rate_fmt,
                            'elapsed': elapsed_str,
                            'remaining': remaining_str,
                            'l_bar': l_bar,
                            'r_bar': r_bar,
                            'desc': prefix if prefix else ''
                            }
                # Interpolate supplied bar format with the dict
                return bar_format.format(**bar_args)

        # no total: no progressbar, ETA, just progress stats
        else:
            return (prefix if prefix else '') + '{0}{1} [{2}, {3}]'.format(
                n_fmt, unit, elapsed_str, rate_fmt)

    def __init__(self, iterable=None, desc=None, total=None, leave=False,
                 file=sys.stderr, ncols=None, mininterval=0.1,
                 maxinterval=10.0, miniters=None, ascii=None,
                 disable=False, unit='it', unit_scale=False,
                 dynamic_ncols=False, smoothing=0.3, nested=False,
                 bar_format=None, initial=0, gui=False):
        """
        Parameters
        ----------
        iterable  : iterable, optional
            Iterable to decorate with a progressbar.
            Leave blank [default: None] to manually manage the updates.
        desc  : str, optional
            Prefix for the progressbar [default: None].
        total  : int, optional
            The number of expected iterations. If not given, len(iterable)
            is used if possible. As a last resort, only basic progress
            statistics are displayed (no ETA, no progressbar). If `gui` is
            True and this parameter needs subsequent updating, specify an
            initial arbitrary large positive integer, e.g. int(9e9).
        leave  : bool, optional
            If [default: False], removes all traces of the progressbar
            upon termination of iteration.
        file  : `io.TextIOWrapper` or `io.StringIO`, optional
            Specifies where to output the progress messages
            [default: sys.stderr]. Uses `file.write(str)` and `file.flush()`
            methods.
        ncols  : int, optional
            The width of the entire output message. If specified,
            dynamically resizes the progressbar to stay within this bound.
            If [default: None], attempts to use environment width. The
            fallback is a meter width of 10 and no limit for the counter and
            statistics. If 0, will not print any meter (only stats).
        mininterval  : float, optional
            Minimum progress update interval, in seconds [default: 0.1].
        maxinterval  : float, optional
            Maximum progress update interval, in seconds [default: 10.0].
        miniters  : int, optional
            Minimum progress update interval, in iterations [default: None].
        ascii  : bool, optional
            If [default: None] or false, use unicode (smooth blocks) to fill
            the meter. The fallback is to use ASCII characters `1-9 #`.
        disable : bool
            Whether to disable the entire progressbar wrapper
            [default: False].
        unit  : str, optional
            String that will be used to define the unit of each iteration
            [default: 'it'].
        unit_scale  : bool, optional
            If set, the number of iterations will be reduced/scaled
            automatically and a metric prefix following the
            International System of Units standard will be added
            (kilo, mega, etc.) [default: False].
        dynamic_ncols  : bool, optional
            If set, constantly alters `ncols` to the environment (allowing
            for window resizes) [default: False].
        smoothing  : float
            Exponential moving average smoothing factor for speed estimates
            (ignored in GUI mode). Ranges from 0 (average speed) to 1
            (current/instantaneous speed) [default: 0.3].
        nested  : bool, optional
            Whether this iterable is nested in another one also managed by
            `tqdm` [default: False]. Allows display of multiple, nested
            progress bars.
        bar_format  : str, optional
            Specify a custom bar string formatting. May impact performance.
            [default: '{l_bar}{bar}{r_bar}'], where l_bar is
            '{desc}{percentage:3.0f}%|' and r_bar is
            '| {n_fmt}/{total_fmt} [{elapsed_str}<{remaining_str}, {rate_fmt}]'.
            Possible vars: bar, n, n_fmt, total, total_fmt, percentage,
            rate, rate_fmt, elapsed, remaining, l_bar, r_bar, desc.
        initial : int, optional
            The initial counter value. Useful when restarting a progress
            bar [default: 0].
        gui  : bool, optional
            WARNING: internal parameter - do not use.
            Use tqdm_gui(...) instead. If set, will attempt to use
            matplotlib animations for a graphical output [default: false].

        Returns
        -------
        out  : decorated iterator.
        """
        # Preprocess the arguments
        if total is None and iterable is not None:
            try:
                total = len(iterable)
            except (TypeError, AttributeError):
                total = None

        if ((ncols is None) and (file in (sys.stderr, sys.stdout))) or \
                dynamic_ncols:
            if dynamic_ncols:  # pragma: no cover
                dynamic_ncols = _environ_cols_wrapper()
                ncols = dynamic_ncols(file)
            else:
                ncols = _environ_cols_wrapper()(file)

        if miniters is None:
            miniters = 0
            dynamic_miniters = True
        else:
            dynamic_miniters = False

        if mininterval is None:
            mininterval = 0

        if maxinterval is None:
            maxinterval = 0

        if ascii is None:
            ascii = not _supports_unicode(file)

        # Convert bar format into unicode since terminal uses unicode
        if bar_format and not ascii:
            bar_format = _unicode(bar_format)

        if smoothing is None:
            smoothing = 0

        # Store the arguments
        self.iterable = iterable
        self.desc = desc + ': ' if desc else ''
        self.total = total
        self.leave = leave
        self.fp = file
        self.ncols = ncols
        self.mininterval = mininterval
        self.maxinterval = maxinterval
        self.miniters = miniters
        self.dynamic_miniters = dynamic_miniters
        self.ascii = ascii
        self.disable = disable
        self.unit = unit
        self.unit_scale = unit_scale
        self.gui = gui
        self.dynamic_ncols = dynamic_ncols
        self.smoothing = smoothing
        self.avg_time = None
        self._time = time
        # if nested, at initial sp() call we replace '\r' by '\n' to
        # not overwrite the outer progress bar
        self.nested = nested
        self.bar_format = bar_format

        # Init the iterations counters
        self.last_print_n = initial
        self.n = initial

        if not gui:
            # Initialize the screen printer
            self.sp = self.status_printer(self.fp)
            if not disable:
                if self.nested:
                    self.fp.write('\n')
                self.sp(self.format_meter(self.n, total, 0,
                        (dynamic_ncols(file) if dynamic_ncols else ncols),
                        self.desc, ascii, unit, unit_scale, None, bar_format))

        # Init the time counter
        self.start_t = self.last_print_t = self._time()

    def __len__(self):
        return len(self.iterable) if self.iterable else self.total

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.close()
        return False

    def __iter__(self):
        ''' Backward-compatibility to use: for x in tqdm(iterable) '''

        # Inlining instance variables as locals (speed optimisation)
        iterable = self.iterable

        # If the bar is disabled, then just walk the iterable
        # (note: keep this check outside the loop for performance)
        if self.disable:
            for obj in iterable:
                yield obj
        else:
            ncols = self.ncols
            mininterval = self.mininterval
            maxinterval = self.maxinterval
            miniters = self.miniters
            dynamic_miniters = self.dynamic_miniters
            unit = self.unit
            unit_scale = self.unit_scale
            ascii = self.ascii
            start_t = self.start_t
            last_print_t = self.last_print_t
            last_print_n = self.last_print_n
            n = self.n
            dynamic_ncols = self.dynamic_ncols
            smoothing = self.smoothing
            avg_time = self.avg_time
            bar_format = self.bar_format
            _time = self._time
            format_meter = self.format_meter

            try:
                sp = self.sp
            except AttributeError:
                raise DeprecationWarning('Please use tqdm_gui(...)'
                                         ' instead of tqdm(..., gui=True)')

            for obj in iterable:
                yield obj
                # Update and print the progressbar.
                # Note: does not call self.update(1) for speed optimisation.
                n += 1
                delta_it = n - last_print_n
                # check the counter first (avoid calls to time())
                if delta_it >= miniters:
                    cur_t = _time()
                    delta_t = cur_t - last_print_t
                    if delta_t >= mininterval:
                        elapsed = cur_t - start_t
                        # EMA (not just overall average)
                        if smoothing:  # and delta_it
                            avg_time = delta_t / delta_it \
                                if avg_time is None \
                                else smoothing * delta_t / delta_it + \
                                (1 - smoothing) * avg_time

                        sp(format_meter(
                            n, self.total, elapsed,
                            (dynamic_ncols(self.fp) if dynamic_ncols
                             else ncols),
                            self.desc, ascii, unit, unit_scale,
                            1 / avg_time if avg_time else None, bar_format))

                        # If no `miniters` was specified, adjust automatically
                        # to the maximum iteration rate seen so far.
                        if dynamic_miniters:
                            if maxinterval and delta_t > maxinterval:
                                # Set miniters to correspond to maxinterval
                                miniters = delta_it * maxinterval / delta_t
                            elif mininterval and delta_t:
                                # EMA-weight miniters to converge
                                # towards the timeframe of mininterval
                                miniters = smoothing * delta_it * mininterval \
                                    / delta_t + (1 - smoothing) * miniters
                            else:
                                miniters = smoothing * delta_it + \
                                    (1 - smoothing) * miniters

                        # Store old values for next call
                        last_print_n = n
                        last_print_t = cur_t

            # Closing the progress bar.
            # Update some internal variables for close().
            self.last_print_n = last_print_n
            self.n = n
            self.close()

    def update(self, n=1):
        """
        Manually update the progress bar, useful for streams
        such as reading files.
        E.g.:
        >>> t = tqdm(total=filesize) # Initialise
        >>> for current_buffer in stream:
        ...    ...
        ...    t.update(len(current_buffer))
        >>> t.close()
        The last line is highly recommended, but possibly not necessary if
        `t.update()` will be called in such a way that `filesize` will be
        exactly reached and printed.

        Parameters
        ----------
        n  : int
            Increment to add to the internal counter of iterations
            [default: 1].
        """
        if self.disable:
            return

        if n < 1:
            raise ValueError("n ({0}) cannot be less than 1".format(n))
        self.n += n

        delta_it = self.n - self.last_print_n  # should be n?
        if delta_it >= self.miniters:
            # We check the counter first, to reduce the overhead of time()
            cur_t = self._time()
            delta_t = cur_t - self.last_print_t
            if delta_t >= self.mininterval:
                elapsed = cur_t - self.start_t
                # EMA (not just overall average)
                if self.smoothing:  # and delta_it
                    self.avg_time = delta_t / delta_it \
                        if self.avg_time is None \
                        else self.smoothing * delta_t / delta_it + \
                        (1 - self.smoothing) * self.avg_time

                if not hasattr(self, "sp"):
                    raise DeprecationWarning('Please use tqdm_gui(...)'
                                             ' instead of tqdm(..., gui=True)')

                self.sp(self.format_meter(
                    self.n, self.total, elapsed,
                    (self.dynamic_ncols(self.fp) if self.dynamic_ncols
                     else self.ncols),
                    self.desc, self.ascii, self.unit, self.unit_scale,
                    1 / self.avg_time if self.avg_time else None,
                    self.bar_format))

                # If no `miniters` was specified, adjust automatically to the
                # maximum iteration rate seen so far.
                # e.g.: After running `tqdm.update(5)`, subsequent
                # calls to `tqdm.update()` will only cause an update after
                # at least 5 more iterations.
                if self.dynamic_miniters:
                    if self.maxinterval and delta_t > self.maxinterval:
                        self.miniters = self.miniters * self.maxinterval \
                            / delta_t
                    elif self.mininterval and delta_t:
                        self.miniters = self.smoothing * delta_it \
                            * self.mininterval / delta_t + \
                            (1 - self.smoothing) * self.miniters
                    else:
                        self.miniters = self.smoothing * delta_it + \
                            (1 - self.smoothing) * self.miniters

                # Store old values for next call
                self.last_print_n = self.n
                self.last_print_t = cur_t

    def close(self):
        """
        Cleanup and (if leave=False) close the progressbar.
        """
        if self.disable:
            return

        if self.leave:
            if self.last_print_n < self.n:
                cur_t = self._time()
                # stats for overall rate (no weighted average)
                self.sp(self.format_meter(
                    self.n, self.total, cur_t - self.start_t,
                    (self.dynamic_ncols(self.fp) if self.dynamic_ncols
                     else self.ncols),
                    self.desc, self.ascii, self.unit, self.unit_scale, None,
                    self.bar_format))
            self.fp.write('\r' + _term_move_up() if self.nested else '\n')
        else:
            self.sp('')  # clear up last bar
            self.fp.write('\r' + _term_move_up() if self.nested else '\r')

    def unpause(self):
        """
        Restart tqdm timer from last print time.
        """
        cur_t = self._time()
        self.start_t += cur_t - self.last_print_t
        self.last_print_t = cur_t

    def set_description(self, desc=None):
        """
        Set/modify description of the progress bar.
        """
        self.desc = desc + ': ' if desc else ''


def trange(*args, **kwargs):
    """
    A shortcut for tqdm(xrange(*args), **kwargs).
    On Python3+ range is used instead of xrange.
    """
    return tqdm(_range(*args), **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import (division, print_function, absolute_import,
                        unicode_literals)

__all__ = ["sample_ball", "MH_proposal_axisaligned", "MPIPool"]


import numpy as np

from .mpi_pool import MPIPool


def sample_ball(p0, std, size=1):
    """
    Produce a ball of walkers around an initial parameter value.

    :param p0: The initial parameter value.
    :param std: The axis-aligned standard deviation.
    :param size: The number of samples to produce.

    """
    assert(len(p0) == len(std))
    return np.vstack([p0 + std * np.random.normal(size=len(p0))
                      for i in range(size)])


class MH_proposal_axisaligned(object):
    """
    A Metropolis-Hastings proposal, with axis-aligned Gaussian steps,
    for convenient use as the ``mh_proposal`` option to
    :func:`EnsembleSampler.sample` .

    """
    def __init__(self, stdev):
        self.stdev = stdev

    def __call__(self, X):
        (nw, npar) = X.shape
        assert(len(self.stdev) == npar)
        return X + self.stdev * np.random.normal(size=X.shape)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import (division, print_function, absolute_import,
                        unicode_literals)

__all__ = ["function", "integrated_time"]

import numpy as np


def function(x, axis=0, fast=False):
    """
    Estimate the autocorrelation function of a time series using the FFT.

    :param x:
        The time series. If multidimensional, set the time axis using the
        ``axis`` keyword argument and the function will be computed for every
        other axis.

    :param axis: (optional)
        The time axis of ``x``. Assumed to be the first axis if not specified.

    :param fast: (optional)
        If ``True``, only use the largest ``2^n`` entries for efficiency.
        (default: False)

    """
    x = np.atleast_1d(x)
    m = [slice(None), ] * len(x.shape)

    # For computational efficiency, crop the chain to the largest power of
    # two if requested.
    if fast:
        n = int(2**np.floor(np.log2(x.shape[axis])))
        m[axis] = slice(0, n)
        x = x
    else:
        n = x.shape[axis]

    # Compute the FFT and then (from that) the auto-correlation function.
    f = np.fft.fft(x-np.mean(x, axis=axis), n=2*n, axis=axis)
    m[axis] = slice(0, n)
    acf = np.fft.ifft(f * np.conjugate(f), axis=axis)[m].real
    m[axis] = 0
    return acf / acf[m]


def integrated_time(x, axis=0, window=50, fast=False):
    """
    Estimate the integrated autocorrelation time of a time series.

    See `Sokal's notes <http://www.stat.unc.edu/faculty/cji/Sokal.pdf>`_ on
    MCMC and sample estimators for autocorrelation times.

    :param x:
        The time series. If multidimensional, set the time axis using the
        ``axis`` keyword argument and the function will be computed for every
        other axis.

    :param axis: (optional)
        The time axis of ``x``. Assumed to be the first axis if not specified.

    :param window: (optional)
        The size of the window to use. (default: 50)

    :param fast: (optional)
        If ``True``, only use the largest ``2^n`` entries for efficiency.
        (default: False)

    """
    # Compute the autocorrelation function.
    f = function(x, axis=axis, fast=fast)

    # Special case 1D for simplicity.
    if len(f.shape) == 1:
        return 1 + 2*np.sum(f[1:window])

    # N-dimensional case.
    m = [slice(None), ] * len(f.shape)
    m[axis] = slice(1, window)
    tau = 1 + 2*np.sum(f[m], axis=axis)

    return tau


if __name__ == "__main__":
    import time
    import acor

    N = 400000
    a = 0.9
    d = 3
    x = np.empty((N, d))
    x[0] = np.zeros(d)
    for i in xrange(1, N):
        x[i] = x[i-1] * a + np.random.rand(d)

    strt = time.time()
    print(integrated_time(x))
    print(time.time() - strt)

    strt = time.time()
    print([acor.acor(x[:, i])[0] for i in range(d)])
    print(time.time() - strt)



"""
Implementation of a class that transforms the input before kicking the arguments
to the higher class. Currently only supports HFSModel.

.. moduleauthor:: Wouter Gins <wouter.gins@kuleuven.be>
"""
from satlas.models.hfsmodel import HFSModel

__all__ = ['TransformHFSModel']

def identity(*args):
    return args[0]

class TransformHFSModel(HFSModel):
    """Create an HFSModel that applies both a pre-processing
    transformation on the input data and a post-processing
    transformation on the output data. Mathematically:

        .. math::
            (post \circ model \circ pre)(input)

    Both transformations are initialized to the identical
    transformation for one input argument."""

    def __init__(self, *args, **kwargs):
        """Passes all arguments on the :class:`.HFSModel`.
        See :class:`.HFSModel` for input information."""
        super(TransformHFSModel, self).__init__(*args, **kwargs)
        self._pre_transform = identity
        self._post_transform = identity

    @property
    def pre_transform(self):
        """The transformation function to be applied to the input data. Wrapping
        with the *functools.lru_cache* function is attempted, and non-callable
        objects raise an error when assigned to :attr:`.pre_transform`."""
        return self._pre_transform

    @pre_transform.setter
    def pre_transform(self, func):
        if callable(func):
            try:
                from functool import lru_cache
                self._pre_transform = lru_cache(maxsize=512, typed=True)(func)
            except:
                self._pre_transform = func
        else:
            raise TypeError('supplied value must be a callable!')

    @property
    def post_transform(self):
        """The transformation function to be applied to the output data.
        Non-callable objects raise an error when assigned to
        :attr:`.post_transform`."""
        return self._post_transform

    @post_transform.setter
    def post_transform(self, func):
        if callable(func):
            self._post_transform = func
        else:
            raise TypeError('supplied value must be a callable!')

    def plot(self, *args, **kwargs):
        """Grants access to the :meth:`.HFSModel.plot` method, passing all
        arguments. The transformation used is temporarily changed to the
        identity transform."""
        # remember_pre = copy.deepcopy(self._pre_transform)
        # remember_post = copy.deepcopy(self._post_transform)
        # self._pre_transform = identity
        to_return = super(TransformHFSModel, self).plot(*args, **kwargs)
        # self._pre_transform = remember_pre
        return to_return

    def __call__(self, *args, **kwargs):
        return self._post_transform(super(TransformHFSModel, self).__call__(self._pre_transform(*args, **kwargs)), *args, **kwargs)



"""
Implementation of base class for extension to models describing actual data.

.. moduleauthor:: Wouter Gins <wouter.gins@kuleuven.be>
.. moduleauthor:: Ruben de Groote <ruben.degroote@kuleuven.be>
"""
import copy

import lmfit as lm
from satlas.loglikelihood import create_gaussian_priormap
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from copy import deepcopy

__all__ = ['load_model']

class SATLASParameters(lm.Parameters):
    _prefix = ''

    def __getitem__(self, key):
        try:
            return super(SATLASParameters, self).__getitem__(key)
        except KeyError:
            try:
                return super(SATLASParameters, self).__getitem__(self._prefix + key)
            except KeyError:
                raise

    def __deepcopy__(self, memo):
        """Parameters deepcopy needs to make sure that
        asteval is available and that all individula
        parameter objects are copied"""
        _pars = SATLASParameters()

        # find the symbols that were added by users, not during construction
        sym_unique = self._asteval.user_defined_symbols()
        unique_symbols = {key: deepcopy(self._asteval.symtable[key], memo)
                              for key in sym_unique}
        _pars._asteval.symtable.update(unique_symbols)

        # we're just about to add a lot of Parameter objects to the newly
        parameter_list = []
        for key, par in self.items():
            if isinstance(par, lm.Parameter):
                param = lm.Parameter(name=par.name,
                                  value=par.value,
                                  min=par.min,
                                  max=par.max)
                param.vary = par.vary
                param.stderr = par.stderr
                param.correl = par.correl
                param.init_value = par.init_value
                param.expr = par.expr
                parameter_list.append(param)

        _pars.add_many(*parameter_list)
        _pars._prefix = self._prefix
        return _pars

class BaseModel(object):

    """Abstract baseclass for all models. For input, see these
    classes."""

    def __init__(self):
        super(BaseModel, self).__init__()
        self._expr = {}
        self._vary = {}
        self._constraints = {}
        self._params = None
        self._lnprior_mapping = {}
        self._chisquare_mapping = {}
        self._prefix = ''

    def _set_prefix(self, value):
        for p in self._parameters:
            if len(self._prefix) > 0:
                if self._parameters[p].expr is not None:
                    self._parameters[p].expr = self._parameters[p].expr.replace(self._prefix, value)
            else:
                if self._parameters[p].expr is not None:
                    for P in self._parameters:
                        if P in self._parameters[p].expr:
                            self._parameters[p].expr = self._parameters[p].expr.replace(P, value + P)
        for p in list(self._parameters.keys()):
            if len(self._prefix) > 0:
                self._parameters[p].name = self._parameters[p].name[len(self._prefix):]
            self._parameters[p].name = value + self._parameters[p].name
            self._parameters[value + p] = self._parameters.pop(p)
        self._parameters._prefix = value
        self._prefix = value

    def _add_prefix(self, value):
        for p in self._parameters:
            if self._parameters[p].expr is not None:
                if len(self._prefix) > 0:
                    self._parameters[p].expr = self._parameters[p].expr.replace(self._prefix, value + self._prefix)
                else:
                    for P in self._parameters:
                        if P in self._parameters[p].expr:
                            self._parameters[p].expr = self._parameters[p].expr.replace(P, value + P)
        for p in list(self._parameters.keys()):
            self._parameters[p].name = value + self._parameters[p].name
            self._parameters[value + p] = self._parameters.pop(p)
        self._parameters._prefix = value + self._parameters._prefix
        self._prefix = value + self._prefix

    def set_literature_values(self, literatureDict):
        """Sets the lnprior and chisquare mapping to handle the given
        literature values and uncertainties.

        Parameters
        ----------
        literatureDict: dictionary
            Dictionary with the parameter names as keys. Each
            key should correspond to a dictionary containing
            a 'value' and 'uncertainty' key."""
        priorDict = {}
        chisquareDict = {}
        for k in literatureDict:
            v, u = literatureDict[k]['value'], literatureDict[k]['uncertainty']

            prior = create_gaussian_priormap(v, u)
            priorDict[k] = prior

            chisquare = lambda value: (value - v) / u
            chisquareDict[k] = chisquare
        self.set_chisquare_mapping(chisquareDict)
        self.set_lnprior_mapping(priorDict)

    def set_lnprior_mapping(self, mappingDict):
        """Sets the prior mapping for the different parameters.
        This will affect likelihood fits.

        Parameters
        ----------
        mappingDict: dictionary
            Dictionary containing the functions that give the
            prior for the given parameter value. Use the parameter
            names as keys."""
        for k in mappingDict.keys():
            self._lnprior_mapping[k] = copy.deepcopy(mappingDict[k])

    def set_chisquare_mapping(self, mappingDict):
        """Sets the prior mapping for the different parameters.
        This will affect chisquare fits.

        Parameters
        ----------
        mappingDict: dictionary
            Dictionary containing the functions that give the
            prior for the given parameter value. Use the parameter
            names as keys."""
        for k in mappingDict.keys():
            self._chisquare_mapping[k] = copy.deepcopy(mappingDict[k])

    def set_value(self, valueDict):
        """Sets the value of the given parameters to the given values.

        Parameters
        ----------
        valueDict: dictionary
            Dictionary containing the values for the parameters
            with the parameter names as keys."""
        p = self.params.copy()
        for key in valueDict:
            try:
                p[key].value = valueDict[key]
            except KeyError:
                pass
        self.params = p.copy()

    def set_expr(self, exprDict):
        """Sets the expression of the selected parameters
        to the given expressions.

        Parameters
        ----------
        exprDict: dictionary
            Dictionary containing the expressions for the parameters
            with the parameter names as keys."""
        p = self.params.copy()
        for k in exprDict.keys():
            self._expr[k] = copy.deepcopy(exprDict[k])
            try:
                p[k].expr = self._expr[k]
                p[k].vary = False
            except KeyError:
                pass
        self.params = p.copy()

    def set_variation(self, varyDict):
        """Sets the variation of the fitparameters as supplied in the
        dictionary.

        Parameters
        ----------
        varyDict: dictionary
            A dictionary containing 'key: True/False' mappings with
            the parameter names as keys."""
        p = self.params.copy()
        for k in varyDict.keys():
            self._vary[k] = copy.deepcopy(varyDict[k])
            try:
                p[k].vary = self._vary[k]
            except KeyError:
                pass
        self.params = p.copy()

    def set_boundaries(self, boundaryDict):
        """Sets the boundaries of the fitparameters as supplied in the
        dictionary.

        Parameters
        ----------
        boundaryDict: dictionary
            A dictionary containing "key: {'min': value, 'max': value}" mappings.
            A missing 'min' or 'max' key gives no boundary
            in that direction. The parameter names have to be used as keys."""
        p = self.params.copy()
        for k in boundaryDict.keys():
            self._constraints[k] = copy.deepcopy(boundaryDict[k])
            for bound in self._constraints[k].keys():
                b = self._constraints[k][bound]
                try:
                    if bound.lower() == 'min':
                        p[k].min = b if b is not None else -np.inf
                    elif bound.lower() == 'max':
                        p[k].max = b if b is not None else np.inf
                    else:
                        pass
                except KeyError:
                    raise
        self.params = p.copy()

    def _check_variation(self, par):
        # Make sure the variations in the params are set correctly.
        for key in self._vary.keys():
            try:
                par[key].vary = self._vary[key]
            except KeyError:
                pass

        for key in self._constraints.keys():
            for bound in self._constraints[key]:
                try:
                    if bound.lower() == 'min':
                        par[key].min = self._constraints[key][bound]
                    elif bound.lower() == 'max':
                        par[key].max = self._constraints[key][bound]
                    else:
                        pass
                except KeyError:
                    pass
        for key in self._expr.keys():
            try:
                par[key].expr = self._expr[self._prefix + key]
            except KeyError:
                pass
        return par.copy()

    def get_chisquare_mapping(self):
        return np.array([self._chisquare_mapping[k](self.params[k].value) for k in self._chisquare_mapping.keys()])

    def get_lnprior_mapping(self, params):
        # Check if the parameter values are within the acceptable range.
        for key in params.keys():
            par = params[key]
            if par.vary:
                try:
                    leftbound, rightbound = (par.priormin,
                                             par.priormax)
                except AttributeError:
                    leftbound, rightbound = par.min, par.max
                leftbound = -np.inf if leftbound is None else leftbound
                rightbound = np.inf if rightbound is None else rightbound
                if not leftbound < par.value < rightbound:
                    return -np.inf
        # If defined, calculate the lnprior for each seperate parameter
        return_value = 1.0
        for key in self._lnprior_mapping.keys():
            return_value += self._lnprior_mapping[key](params[key].value)
        return return_value

    def display_mle_fit(self, scaled=False, **kwargs):
        """Give a readable overview of the result of the MLE fitting routine.

        Warning
        -------
        The uncertainty shown is the largest of the asymmetrical errors! Work
        is being done to incorporate asymmetrical errors in the report; for
        now, rely on the correlation plot."""
        if hasattr(self, 'fit_mle'):
            if 'show_correl' not in kwargs:
                kwargs['show_correl'] = False
            print('NDoF: {:d}, Chisquare: {:.8G}, Reduced Chisquare: {:.8G}'.format(self.ndof_mle, self.chisqr_mle, self.redchi_mle))
            if scaled:
                print('Errors scaled with reduced chisquare.')
                par = copy.deepcopy(self.fit_mle)
                for p in par:
                    if par[p].stderr is not None:
                        par[p].stderr *= (self.redchi_mle**0.5)
                print(lm.fit_report(par, **kwargs))
            else:
                print('Errors not scaled with reduced chisquare.')
                print(lm.fit_report(self.fit_mle, **kwargs))
        else:
            print('Model has not yet been fitted with this method!')

    def display_chisquare_fit(self, scaled=False, **kwargs):
        """Display all relevent info of the least-squares fitting routine,
        if this has been performed.

        Parameters
        ----------
        kwargs: misc
            Keywords passed on to :func:`fit_report` from the LMFit package."""
        if hasattr(self, 'chisq_res_par'):
            print('NDoF: {:d}, Chisquare: {:.8G}, Reduced Chisquare: {:.8G}'.format(self.ndof_chi, self.chisqr_chi, self.redchi_chi))
            print('Akaike Information Criterium: {:.8G}, Bayesian Information Criterium: {:.8G}'.format(self.aic_chi, self.bic_chi))
            if scaled:
                print('Errors scaled with reduced chisquare.')
                par = copy.deepcopy(self.chisq_res_par)
                for p in par:
                    if par[p].stderr is not None:
                        par[p].stderr *= (self.redchi_chi**0.5)
                print(lm.fit_report(par, **kwargs))
            else:
                print('Errors not scaled with reduced chisquare.')
                print(lm.fit_report(self.chisq_res_par, **kwargs))
        else:
            print('Spectrum has not yet been fitted with this method!')

    def get_goodness_of_fit(self, selection='chisquare'):
        if selection.lower() == 'chisquare':
            return (self.ndof_chi, self.chisqr_chi, self.redchi_chi, self.aic_chi, self.bic_chi)
        elif selection.lower() == 'mle':
            return (self.ndof_mle, self.chisqr_mle, self.redchi_mle, self.aic_mle, self.bic_mle)


    def get_result(self, selection='any'):
        """Return the variable names, values and estimated error bars for the
        parameters as seperate lists.

        Parameters
        ----------
        selection: string, optional
            Selects if the chisquare ('chisquare' or 'any') or MLE values are
            used. Defaults to 'any'.

        Returns
        -------
        names, values, uncertainties: tuple of lists
            Returns a 3-tuple of lists containing the names of the parameters,
            the values and the estimated uncertainties."""
        var, var_names, varerr = [], [], []
        if hasattr(self, 'chisq_res_par') and (selection.lower() == 'chisquare'
                                               or selection.lower() == 'any'):
            for key in sorted(self.chisq_res_par.keys()):
                if self.chisq_res_par[key].vary:
                    var.append(self.chisq_res_par[key].value)
                    var_names.append(self.chisq_res_par[key].name)
                    varerr.append(self.chisq_res_par[key].stderr)
        elif hasattr(self, 'fit_mle'):
            for key in sorted(self.fit_mle.params.keys()):
                if self.fit_mle.params[key].vary:
                    var.append(self.fit_mle.params[key].value)
                    var_names.append(self.fit_mle.params[key].name)
                    varerr.append(self.fit_mle.params[key].stderr)
        else:
            params = self.params
            for key in sorted(params.keys()):
                if params[key].vary:
                    var.append(params[key].value)
                    var_names.append(params[key].name)
                    varerr.append(None)
        return var_names, var, varerr

    def get_result_frame(self, method='chisquare', selected=False, bounds=False, vary=False, scaled=False):
        """Returns the data from the fit in a pandas DataFrame.

        Parameters
        ----------
        method: str, optional
            Selects which fitresults have to be loaded. Can be 'chisquare' or
            'mle'. Defaults to 'chisquare'.
        selected: list of strings, optional
            Selects the parameters that have any string in the list
            as a substring in their name. Set to *None* to select
            all parameters. Defaults to *None*.
        bounds: boolean, optional
            Selects if the boundary also has to be given. Defaults to
            *False*.
        vary: boolean, optional
            Selects if only the parameters that have been varied have to
            be supplied. Defaults to *False*.
        scaled: boolean, optional
            Sets the uncertainty scaling with the reduced chisquare value. Default to *False*.

        Returns
        -------
        resultframe: DataFrame
            Dateframe with MultiIndex, using the variable names as main column names
            and either two subcolumns for the value and the uncertainty, or
            four subcolumns for the value, uncertainty and bounds."""
        if method.lower() == 'chisquare':
            if scaled:
                p = copy.deepcopy(self.chisq_res_par)
                for par in p:
                    p[par].stderr *= self.redchi_chi**0.5
            else:
                p = self.chisq_res_par
        elif method.lower() == 'mle':
            if scaled:
                p = copy.deepcopy(self.fit_mle)
                for par in p:
                    p[par].stderr *= self.redchi_mle**0.5
            else:
                p = self.fit_mle
        values = p.values()
        if selected:
            values = [v for n in self.selected for v in values if n in v.name]
        if vary:
            values = [v for v in values if v.vary]
        if bounds:
            ind = ['Value', 'Uncertainty', 'Upper Bound', 'Lower Bound']
            data = np.array([[p.value, p.stderr, p.max, p.min] for p in values]).flatten()
            columns = [[p.name for pair in zip(values, values) for p in pair],
                       [x for p in values for x in ind]]
        else:
            data = np.array([[p.value, p.stderr] for p in values]).flatten()
            ind = ['Value', 'Uncertainty']
            columns = [[p.name for pair in zip(values, values) for p in pair],
                       [x for p in values for x in ind]]
        columns = pd.MultiIndex.from_tuples(list(zip(*columns)))
        result = pd.DataFrame(data, index=columns).T
        if method.lower() == 'chisquare':
            result.loc[:, 'Chisquare'] = pd.Series(np.array([self.chisqr_chi]), index=result.index)
            result.loc[:, 'Reduced chisquare'] = pd.Series(np.array([self.redchi_chi]), index=result.index)
            result.loc[:, 'NDoF'] = pd.Series(np.array([self.ndof_chi]), index=result.index)
        else:
            result.loc[:, 'Chisquare'] = pd.Series(np.array([self.chisqr_mle]), index=result.index)
            result.loc[:, 'Reduced chisquare'] = pd.Series(np.array([self.redchi_mle]), index=result.index)
            result.loc[:, 'NDoF'] = pd.Series(np.array([self.ndof_mle]), index=result.index)
        return result

    def get_result_dict(self, method='chisquare', scaled=False):
        """Returns the fitted parameters in a dictionary of the form {name: [value, uncertainty]}.

        Parameters
        ----------
        method: {'chisquare', 'mle'}
            Selects which parameters have to be returned.
        scaled: boolean
            Selects if, in case of chisquare parameters, the uncertainty
            has to be scaled by sqrt(reduced_chisquare). Defaults to *False*.

        Returns
        -------
        dict
            Dictionary of the form described above."""
        if method.lower() == 'chisquare':
            if scaled:
                p = copy.deepcopy(self.chisq_res_par)
                for par in p:
                    p[par].stderr *= self.redchi_chi**0.5
            else:
                p = self.chisq_res_par
        else:
            if scaled:
                p = copy.deepcopy(self.fit_mle)
                for par in p:
                    p[par].stderr *= self.redchi_mle**0.5
            else:
                p = self.fit_mle
        returnDict = {P: [p[P].value, p[P].stderr] for P in p}
        return returnDict

    def save(self, path):
        """Saves the current spectrum, including the results of the fitting
        and the parameters, to the specified file.

        Parameters
        ----------
        path: string
            Name of the file to be created."""
        import pickle
        with open(path, 'wb') as f:
            pickle.dump(self, f)

    #######################################
    #      METHODS CALLED BY FITTING      #
    #######################################

    def _sanitize_input(self, x, y, yerr=None):
        return x, y, yerr

    def seperate_response(self, x):
        """Wraps the output of the :meth:`__call__` in a list, for
        ease of coding in the fitting routines."""
        return [self(x)]

    def __add__(self, other):
        """Add two spectra together to get an :class:`.SumModel`.

        Parameters
        ----------
        other: BaseModel
            Other spectrum to add.

        Returns
        -------
        SumModel
            A SumModel combining both spectra."""
        from .summodel import SumModel
        if isinstance(other, SumModel):
            l = [self] + other.models
        else:
            l = [self, other]
        return SumModel(l)

    def __radd__(self, other):
        if other == 0:
            return self
        else:
            return self.__add__(other)

    def __call__(self, x):
        raise NotImplementedError("Method has to be implemented in subclass!")

def load_model(path):
    """Loads the saved BaseModel and returns the reconstructed object.

    Parameters
    ----------
    path: string
        Location of the saved model.

    Returns
    -------
    model: BaseModel
        Saved BaseModel/child class instance."""
    import pickle
    with open(path, 'rb') as f:
        model = pickle.load(f)
    try:
        for m in model.models:
            m._parameters._prefix = m._prefix
    except:
        model._parameters._prefix = model._prefix
    return model
